import logging
import tempfile
from typing import Any, Optional, Tuple

from google.cloud import exceptions, storage  # type: ignore

from whylogs.api.writer import Writer
from whylogs.api.writer.writer import Writable
from whylogs.core.utils import deprecated_alias

logger = logging.getLogger(__name__)


class GCSWriter(Writer):
    """
    A whylogs writer to upload DatasetProfileViews onto Google Cloud Storage (GCS).

    >**IMPORTANT**: In order to correctly connect to your GCS container, make sure you have
    the following environment variables set: `[GOOGLE_APPLICATION_CREDENTIALS]`

    Parameters
    ----------
    bucket_name = "your-bucket-name"
    source_file_name = "local/path/to/file"
    destination_blob_name = "storage-object-name"

    Returns
    -------
        None

    Examples
    --------
    An example usage of this method can be represented with the simple code above. Here we
    are assuming the user has already assigned a pandas.DataFrame to the `df` variable.

    ```python
    import whylogs as why

    profile = why.log(pandas=df)
    profile.writer("gcs").option(bucket_name="my_bucket").write()
    ```
    """

    def __init__(
        self,
        gcs_client: Optional[storage.Client] = None,
        object_name: Optional[str] = None,
        bucket_name: Optional[str] = None,
    ):
        self.gcs_client = gcs_client or storage.Client()
        self.object_name = object_name
        self.bucket_name = bucket_name

    @deprecated_alias(profile="file")
    def write(
        self,
        file: Writable,
        dest: Optional[str] = None,
        **kwargs: Any,
    ) -> Tuple[bool, str]:
        dest = dest or file.get_default_path()  # type: ignore
        if self.object_name is None:
            self.object_name = dest
        bucket = self.gcs_client.bucket(self.bucket_name)
        blob = bucket.blob(self.object_name)
        try:
            with tempfile.NamedTemporaryFile() as tmp_file:
                file.write(path=tmp_file.name)  # type: ignore
                tmp_file.flush()
                blob.upload_from_filename(tmp_file.name)
        except exceptions.Forbidden as e:
            logging.error(e)
            return False, str(e)
        return True, f"Uploaded {tmp_file.name} to {self.bucket_name}/{self.object_name}"

    def option(
        self,
        bucket_name: Optional[str] = None,
        object_name: Optional[str] = None,
        gcs_client: Optional[Any] = None,
    ) -> "GCSWriter":  # type: ignore
        if bucket_name:
            self.bucket_name = bucket_name
        if object_name:
            self.object_name = object_name
        if gcs_client:
            self.gcs_client = gcs_client
        return self



from typing import Callable, Generic, Optional, TypeVar

T = TypeVar("T")


class Lazy(Generic[T]):
    def __init__(self, fn: Callable[[], T]) -> None:
        self.fn = fn
        self.__value: Optional[T] = None

    @property
    def value(self) -> T:
        if self.__value is None:
            self.__value = self.fn()
        return self.__value



import getpass
import sys
from typing import List, Optional

from whylogs.api.whylabs.session.session_types import ApiKey
from whylogs.api.whylabs.session.session_types import InteractiveLogger as il
from whylogs.api.whylabs.session.session_types import (
    SessionType,
    parse_api_key,
    validate_org_id,
)


def _get_user_choice(prompt: str, options: List[str]) -> int:
    il.question(prompt, ignore_suppress=True)
    for i, option in enumerate(options, 1):
        il.option(f"{i}. {option}", ignore_suppress=True)

    while True:
        try:
            sys.stdout.flush()
            il.message(ignore_suppress=True)
            choice = int(input("Enter a number from the list: "))
            if 1 <= choice <= len(options):
                return choice
        except ValueError:
            pass


def prompt_session_type(allow_anonymous: bool = True, allow_local: bool = False) -> SessionType:
    options = ["WhyLabs. Use an api key to upload to WhyLabs."]

    if allow_anonymous:
        options.append("WhyLabs Anonymous. Upload data anonymously to WhyLabs and get a viewing url.")

    if allow_local:
        options.append("Local. Don't upload data anywhere.")

    if len(options) == 1:
        return SessionType.WHYLABS

    choice = _get_user_choice("What kind of session do you want to use?", options)
    return [SessionType.WHYLABS, SessionType.WHYLABS_ANONYMOUS, SessionType.LOCAL][choice - 1]


def prompt_default_dataset_id() -> Optional[str]:
    try:
        sys.stdout.flush()
        il.message(ignore_suppress=True)
        default_dataset_id = input("[OPTIONAL] Enter a default dataset id to upload to: ").strip()
        return default_dataset_id
    except Exception:
        return None


def prompt_api_key() -> ApiKey:
    while True:
        try:
            sys.stdout.flush()
            il.message(ignore_suppress=True)
            api_key = getpass.getpass(
                "Enter your WhyLabs api key. You can find it at https://hub.whylabsapp.com/settings/access-tokens: "
            )
            return parse_api_key(api_key)
        except Exception:
            il.warning(
                f"Couldn't parse the api key. Expected a key with the format 'key_id.key:org_id'. Got: {api_key}",
                ignore_suppress=True,
            )


def prompt_org_id() -> str:
    while True:
        try:
            sys.stdout.flush()
            il.message(ignore_suppress=True)
            org_id = input("Enter your org id. You can find it at https://hub.whylabsapp.com/settings/access-tokens: ")
            validate_org_id(org_id)
            return org_id
        except Exception:
            il.warning(
                f"Couldn't parse the org id. Expected an id that starts with 'org-'. Got: {org_id}",
                ignore_suppress=True,
            )



from typing import Any, Tuple

from fugue import DataFrame, DataFrames, Outputter
from fugue.extensions import namespace_candidate
from fugue.plugins import parse_outputter
from IPython.display import display

from whylogs.api.fugue import fugue_profile
from whylogs.viz import NotebookProfileVisualizer

from ..logger import log


class WhyViz(Outputter):
    def __init__(self, func: str, **kwargs: Any) -> None:
        super().__init__()
        assert func in ["viz"]
        self._func = func
        self._args = kwargs

    def validate_on_compile(self) -> None:
        if len(self.partition_spec.partition_by) > 0:
            raise NotImplementedError("prepartition by is not supported for visualization")

    def process(self, dfs: DataFrames) -> None:
        visualization = NotebookProfileVisualizer()
        if len(dfs) == 1:
            assert len(dfs) == 1, "viz_profile can take only one dataframe"
            p = self._profile(dfs[0])
            visualization.set_profiles(target_profile_view=p)
            display(visualization.profile_summary())
        elif len(dfs) == 2 and dfs.has_key:
            ref = self._profile(dfs["reference"])
            target = self._profile(dfs["target"])
            visualization.set_profiles(target_profile_view=target, reference_profile_view=ref)
            display(visualization.summary_drift_report())
        else:
            raise ValueError(self._func + " does not support the input")

    def _profile(self, df: DataFrame) -> Any:
        if self.execution_engine.is_distributed:
            partition = None if self.partition_spec.empty else self.partition_spec
            return fugue_profile(df, engine=self.execution_engine, partition=partition, **self._args)
        else:
            return log(df.as_pandas()).view()


@parse_outputter.candidate(namespace_candidate("why", lambda x: isinstance(x, str)))
def _parse_why_outputter(obj: Tuple[str, str], **kwargs: Any) -> Outputter:
    return WhyViz(obj[1], **kwargs)



from whylogs.core.configs import SummaryConfig
from whylogs.core.constraints import (
    DatasetConstraint,
    MetricConstraint,
    PrefixCondition,
)


def column_is_probably_unique(column_name: str, hll_stddev: int = 3) -> MetricConstraint:
    is_probably_unique_constraint = DatasetConstraint(
        name=f"{column_name} is probably unique",
        condition=PrefixCondition(
            f"and <= :{column_name}:cardinality/lower_{hll_stddev} - :{column_name}:counts/n :{column_name}:counts/null  <= - :{column_name}:counts/n :{column_name}:counts/null :{column_name}:cardinality/upper_{hll_stddev}",  # noqa: E501
            cfg=SummaryConfig(hll_stddev=hll_stddev),
        ),
    )
    return is_probably_unique_constraint



import json
import logging
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from whylabs_client import ApiClient
from whylabs_client.api.debug_events_api import DebugEventsApi

from whylogs.api.whylabs.session.session_manager import _default_init
from whylogs.api.whylabs.session.whylabs_client_cache import (
    ClientCacheConfig,
    WhylabsClientCache,
)
from whylogs.core.metadata import _populate_common_profile_metadata
from whylogs.core.utils import ensure_timezone
from whylogs.core.utils.timestamp_calculations import to_utc_milliseconds

from .file_name import _get_event_file_name

diagnostic_logger = logging.getLogger(__name__)


class DebugClient:
    def __init__(
        self,
        api_client: Optional[ApiClient] = None,
        org_id: Optional[str] = None,
        dataset_id: Optional[str] = None,
        ssl_ca_cert: Optional[str] = None,
    ):
        if api_client is None:
            session = _default_init()
            config = session.config
            self._cache_config = ClientCacheConfig(
                ssl_ca_cert=ssl_ca_cert,
                api_key=config.require_api_key(),
            )
            cache = WhylabsClientCache.instance()
            self._api_client, _ = cache.get_client(self._cache_config)
            self._org_id = config.require_org_id()
            self._dataset_id = config.require_default_dataset_id()
        else:
            self._api_client = api_client
            self._org_id = org_id
            self._dataset_id = dataset_id
        self._debug_events_api = DebugEventsApi(self._api_client)

    def log_debug_event(
        self,
        debug_event: Optional[Dict[str, Any]] = None,
        *,
        trace_id: str,
        tags: Optional[List[str]] = None,
        dataset_timestamp: Optional[datetime] = None,
        segment_key_values: Optional[Dict[str, str]] = None,
    ):
        from whylabs_client.model.debug_event import DebugEvent
        from whylabs_client.model.segment import Segment
        from whylabs_client.model.segment_tag import SegmentTag

        if segment_key_values is not None and not isinstance(segment_key_values, Dict):
            diagnostic_logger.warning(
                f"segment_key_values should be a dictionary of key value pairs but was passed as type: {type(segment_key_values)}"
            )

        segment_tags = list()
        now_ms = to_utc_milliseconds(datetime.now(timezone.utc))
        if segment_key_values is not None:
            for segment_key, segment_value in segment_key_values.items():
                if not isinstance(segment_value, str):
                    diagnostic_logger.info(
                        "debug event segment_key_values has a non-string value for "
                        f"{segment_key}: {segment_value}, converting to str"
                    )
                    segment_value = str(segment_value)
                segment_tags.append(SegmentTag(key=segment_key, value=segment_value))
        debug_segment = Segment(tags=segment_tags)
        if dataset_timestamp is not None:
            ensure_timezone(dataset_timestamp)
        dataset_timestamp_in_ms = to_utc_milliseconds(dataset_timestamp) if dataset_timestamp else now_ms
        whylabs_debug_event = DebugEvent(
            content=json.dumps(debug_event),
            trace_id=trace_id,
            tags=tags,
            segment=debug_segment,
            dataset_timestamp=dataset_timestamp_in_ms,
            creation_timestamp=now_ms,
        )
        # TODO: retry
        try:
            status = self._debug_events_api.log_debug_event(
                org_id=self._org_id, dataset_id=self._dataset_id, debug_event=whylabs_debug_event
            )
            diagnostic_logger.info(
                f"Wrote debug event to WhyLabs with trace_id: {trace_id} and content {debug_event}"
                f" status is ({status})"
            )
            return (True, trace_id)
        except Exception as e:
            diagnostic_logger.exception(
                f"Failed to write debug event to {self._org_id}/{self._dataset_id} for trace_id: ("
                f"{trace_id}) error was: {e}"
            )
            return (False, f"Failed writing event with trace_id: ({trace_id}) error was: {e}")


def log_debug_event(
    debug_event: Optional[Dict[str, Any]] = None,
    *,
    trace_id: str,
    name: Optional[str] = None,
    tags: Optional[List[str]] = None,
    dataset_timestamp: Optional[datetime] = None,
    segment_key_values: Optional[Dict[str, str]] = None,
    write_local_file: bool = False,
):
    default_results = (False, f"Error writing debug event: {debug_event}")
    file_results = None
    api_client_results = None
    if write_local_file:
        filename = _get_event_file_name(trace_id=trace_id, timestamp=dataset_timestamp)
        extracted_event = _populate_common_profile_metadata(
            metadata=debug_event, name=name, trace_id=trace_id, tags=tags, timestamp=dataset_timestamp
        )
        with open(filename, "w") as json_file:
            json.dump(extracted_event, json_file)

        diagnostic_logger.info(f"debug_event saved locally to {filename}")
        file_results = (True, filename)
    else:
        debug_writer = DebugClient()
        api_client_results = debug_writer.log_debug_event(
            debug_event=debug_event,
            trace_id=trace_id,
            tags=tags,
            dataset_timestamp=dataset_timestamp,
            segment_key_values=segment_key_values,
        )

    results = api_client_results or file_results or default_results
    return results



import logging
import queue
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, List, Optional, Type, TypeVar, Union

from whylogs.api.logger.experimental.logger.actor.list_util import type_batched_items

QueueMessageType = TypeVar("QueueMessageType")
_DEFAULT_TIMEOUT = 0.1


class QueueWrapper(Generic[QueueMessageType]):
    @abstractmethod
    def send(self, message: QueueMessageType, timeout: float = _DEFAULT_TIMEOUT) -> None:
        raise NotImplementedError()

    @abstractmethod
    def send_many(self, message: QueueMessageType, timeout: float = _DEFAULT_TIMEOUT) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get(self, timeout: float = _DEFAULT_TIMEOUT) -> Optional[QueueMessageType]:
        raise NotImplementedError()

    @abstractmethod
    def get_many(self, timeout: float = _DEFAULT_TIMEOUT, max: Optional[int] = None) -> List[QueueMessageType]:
        raise NotImplementedError()

    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError()


class CloseMessage:
    pass


@dataclass(frozen=True)
class QueueConfig:
    """
    Configuration for the queue used by the actor.

    Parameters
    ----------
    max_batch_size : int
        Maximum number of messages to process in a single batch.
    message_accumualtion_duration : float
        How long to accumulate messages before processing a batch
    message_poll_wait : float
        How long to block when polling for messages in a loop.
    """

    max_batch_size: int = 50_000
    message_accumualtion_duration: float = 1.0  # seconds
    message_poll_wait: float = 0.1  # seconds
    max_buffer_bytes: int = 100_000_000  # 100 MB


MessageType = TypeVar("MessageType")
Messages = Union[MessageType, CloseMessage]


class Actor(ABC, Generic[MessageType]):
    """
    Simple actor inspired interface designed for batch processing messages.
    This was created with thread and process based whylogs logging in mind and it has
    all of the common message passing logic needed.
    """

    def __init__(
        self,
        queue_wrapper: QueueWrapper[Messages],
        queue_config: QueueConfig = QueueConfig(),
    ) -> None:
        self._queue: QueueWrapper[Messages] = queue_wrapper
        self._logger = logging.getLogger(f"ai.whylabs.actor.{type(self).__name__}")
        self._logger.setLevel(logging.DEBUG)
        self._queue_config = queue_config
        super().__init__()

    @abstractmethod
    def close_message_handled(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def set_close_message_handled(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def close_message_wait(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def is_done(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def done_wait(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def set_done(self) -> None:
        """
        Set this actor as done, meaning it has finished processing all messages.
        """
        raise NotImplementedError()

    @abstractmethod
    def is_closed(self) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def set_closed(self) -> None:
        """
        Sets this actor as closed, meaning it should no longer accept messages.
        """
        raise NotImplementedError()

    @abstractmethod
    def process_batch(self, batch: List[Messages], batch_type: Type) -> None:
        raise NotImplementedError()

    def send(self, message: Messages) -> None:
        if self.is_closed():
            raise Exception("Actor is closed, can't send message.")

        if isinstance(message, CloseMessage):
            self.set_closed()

        done = False
        while not done:
            try:
                self._queue.send(message, timeout=self._queue_config.message_poll_wait)
                done = True
            except queue.Full:
                current_size = self._queue.size()
                if current_size == 0:
                    str_msg = str(message)
                    msg = str_msg[: min(len(str_msg), 400)]
                    self._logger.error(
                        f"Message queue can't fit '{msg}...', dropping it permanently. Message was too big to ever fit."
                    )
                    done = True
                else:
                    self._logger.warning(f"Message queue full, trying again. Current size :{self._queue.size()}")
            except Exception as e:
                self._logger.exception(e)

    def send_many(self, messages: List[Messages]) -> None:
        self._queue.send_many(messages)

    def close(self) -> None:
        self._logger.info("Sending Close message.")
        self.send(CloseMessage())
        self._logger.info(f"Shutdown after all pending {self._queue.size()} messages have been processed.")
        self.done_wait()

    def _polling_condition(self, batch_len: int, max: int, last_message_time: float, remaining: int) -> bool:
        if self.is_done() and remaining == 0:
            self._logger.info("Stopping poll. Handled all messages and shutting down.")
            return False

        if batch_len >= max:
            self._logger.info(f"Stopping poll. Got {batch_len} messages.")
            return False

        if time.perf_counter() - last_message_time > self._queue_config.message_accumualtion_duration:
            return False

        return True

    def _load_messages(self) -> Optional[List[Messages]]:
        max = self._queue_config.max_batch_size
        batch: List[Messages] = []
        last_message_time = time.perf_counter()

        while self._polling_condition(len(batch), max, last_message_time, self._queue.size()):
            try:
                batch += self._queue.get_many(timeout=self._queue_config.message_poll_wait, max=max)
                self._logger.info(f"Got {len(batch)} messages. {self._queue.size()} remaining")
            except queue.Empty:
                if self.is_closed() and self.close_message_handled():
                    self._logger.info("Queue closed and no more messages to process.")
                    return None if batch == [] else batch

        return batch

    def process_messages(self) -> None:
        messages: Optional[List[Messages]] = []
        while messages is not None:
            messages = self._load_messages()

            if messages is None:
                continue

            for batch, batch_type in type_batched_items(messages):
                if batch == []:
                    continue

                self._logger.info(
                    f"Processing batch of {len(batch)} {batch_type.__name__}. {self._queue.size()} remaining"
                )

                try:
                    self.process_batch(batch, batch_type)

                    if batch_type == CloseMessage:
                        self.set_close_message_handled()

                except Exception as e:
                    self._logger.error("Error while processing messages")
                    self._logger.exception(e)

        # Can only get here if we're done processing messages
        self.set_done()

    def run(self) -> None:
        try:
            self.process_messages()
        except Exception as e:
            self._logger.error("Error while in main processing loop")
            self._logger.exception(e)
        finally:
            self._logger.info("Shutting down.")



import logging
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)

from whylogs.core.datatypes import DataType, StandardTypeMapper, TypeMapper
from whylogs.core.metrics.metrics import Metric, MetricConfig
from whylogs.core.resolvers import NO_FI_RESOLVER, MetricSpec, ResolverSpec
from whylogs.core.schema import DeclarativeSchema
from whylogs.core.segmentation_partition import SegmentationPartition
from whylogs.core.stubs import pd
from whylogs.core.validators.validator import Validator
from whylogs.experimental.core.metrics.udf_metric import (
    _reset_metric_udfs,
    generate_udf_resolvers,
)
from whylogs.experimental.core.validators.validator import generate_validators

logger = logging.getLogger(__name__)


@dataclass
class UdfSpec:
    """
    Defines UDFs to apply to matching input columns.

    For UDFs matched by column_name(s), the function is passed a dictionary or
    dataframe with the named columns available (the UDF will not be called unless
    all the named columns are available). The output column name is the key in
    the udfs dictionary.

    For UDFs matched by column_type, the function is passed the value or Pandas
    series. The output column name is the key in the udfs dictionary prefixed
    by the input column name.

    You must specify exactly one of column_names or column_type.
    """

    column_names: Optional[List[str]] = None
    udfs: Dict[str, Callable[[Any], Any]] = field(
        default_factory=dict
    )  # new column name -> callable to compute new column value
    column_type: Optional[DataType] = None

    def __post_init__(self):
        if self.column_type is not None:
            if self.column_names:
                raise ValueError("UdfSpec cannot specify both column_names and column_type")
        elif self.column_names is None:
            raise ValueError("UdfSpec must specify column_names or column_type")
        elif len(self.column_names) == 0 or not all([isinstance(x, str) for x in self.column_names]):
            raise ValueError("UdfSpec column_names must be a non-empty list of strings")


def _apply_udfs_on_row(
    values: Union[List, Dict[str, List]], udfs: Dict, new_columns: Dict[str, Any], input_cols: Collection[str]
) -> None:
    for new_col, udf in udfs.items():
        if new_col in input_cols:
            continue

        try:
            new_columns[new_col] = udf(values)[0]
        except Exception:  # noqa
            new_columns[new_col] = None
            logger.exception(f"Evaluating UDF {new_col} failed")


def _apply_udfs_on_dataframe(
    pandas: pd.DataFrame, udfs: Dict, new_df: pd.DataFrame, input_cols: Collection[str]
) -> None:
    for new_col, udf in udfs.items():
        if new_col in input_cols:
            continue

        try:
            new_df[new_col] = pd.Series(udf(pandas))
        except Exception as e:  # noqa
            new_df[new_col] = pd.Series([None])
            logger.exception(f"Evaluating UDF {new_col} failed on columns {pandas.keys()} with error {e}")


def _apply_type_udfs(pandas: pd.Series, udfs: Dict, new_df: pd.DataFrame, input_cols: Collection[str]) -> None:
    for new_col, udf in udfs.items():
        if new_col in input_cols:
            continue

        try:
            new_df[new_col] = pd.Series(udf(pandas))
        except Exception as e:  # noqa
            new_df[new_col] = pd.Series([None])
            logger.exception(f"Evaluating UDF {new_col} failed on column {new_col} with error {e}")


class UdfSchema(DeclarativeSchema):
    """
    Subclass of DeclarativeSchema that runs the UDFs specified in udf_specs to
    create new columns before resolving metrics.
    """

    def __init__(
        self,
        resolvers: Optional[List[ResolverSpec]] = None,
        types: Optional[Dict[str, Any]] = None,
        default_config: Optional[MetricConfig] = None,
        type_mapper: Optional[TypeMapper] = None,
        cache_size: int = 1024,
        schema_based_automerge: bool = False,
        segments: Optional[Dict[str, SegmentationPartition]] = None,
        validators: Optional[Dict[str, List[Validator]]] = None,
        udf_specs: Optional[List[UdfSpec]] = None,
    ) -> None:
        super().__init__(
            resolvers=resolvers,
            types=types,
            default_config=default_config,
            type_mapper=type_mapper,
            cache_size=cache_size,
            schema_based_automerge=schema_based_automerge,
            segments=segments,
            validators=validators,
        )
        udf_specs = udf_specs if udf_specs else []
        self.multicolumn_udfs = [spec for spec in udf_specs if spec.column_names]
        self.type_udfs = defaultdict(list)
        for spec in udf_specs:
            if spec.column_type:
                self.type_udfs[spec.column_type].append(spec)

    def copy(self) -> "UdfSchema":
        copy = super().copy()
        copy.multicolumn_udfs = deepcopy(self.multicolumn_udfs)
        copy.type_udfs = deepcopy(self.type_udfs)
        return copy

    def _run_udfs_on_row(
        self, row: Mapping[str, Any], new_columns: Dict[str, Any], input_cols: Collection[str]
    ) -> None:
        for spec in self.multicolumn_udfs:
            if spec.column_names and set(spec.column_names).issubset(set(row.keys())):
                inputs = {col: [row[col]] for col in spec.column_names}
                _apply_udfs_on_row(inputs, spec.udfs, new_columns, input_cols)

        for column, value in row.items():
            why_type = type(self.type_mapper(type(value)))
            for spec in self.type_udfs[why_type]:
                udfs = {f"{column}.{key}": spec.udfs[key] for key in spec.udfs.keys()}
                _apply_udfs_on_row([value], udfs, new_columns, input_cols)

    def _run_udfs_on_dataframe(self, pandas: pd.DataFrame, new_df: pd.DataFrame, input_cols: Collection[str]) -> None:
        for spec in self.multicolumn_udfs:
            if spec.column_names and set(spec.column_names).issubset(set(pandas.keys())):
                _apply_udfs_on_dataframe(pandas[spec.column_names], spec.udfs, new_df, input_cols)

        for column, dtype in pandas.dtypes.items():
            why_type = type(self.type_mapper(dtype))
            for spec in self.type_udfs[why_type]:
                udfs = {f"{column}.{key}": spec.udfs[key] for key in spec.udfs.keys()}
                _apply_type_udfs(pandas[column], udfs, new_df, input_cols)

    def _run_udfs(
        self, pandas: Optional[pd.DataFrame] = None, row: Optional[Dict[str, Any]] = None
    ) -> Tuple[Optional[pd.DataFrame], Optional[Mapping[str, Any]]]:
        new_columns = deepcopy(row) if row else None
        new_df = pd.DataFrame() if pandas is not None else None
        if row is not None:
            self._run_udfs_on_row(row, new_columns, row.keys())  # type: ignore

        if pandas is not None:
            self._run_udfs_on_dataframe(pandas, new_df, pandas.keys())
            new_df = pd.concat([pandas, new_df], axis=1)

        return new_df, new_columns

    def apply_udfs(
        self, pandas: Optional[pd.DataFrame] = None, row: Optional[Dict[str, Any]] = None
    ) -> Tuple[Optional[pd.DataFrame], Optional[Mapping[str, Any]]]:
        return self._run_udfs(pandas, row)


_multicolumn_udfs: Dict[str, List[UdfSpec]] = defaultdict(list)
_resolver_specs: Dict[str, List[ResolverSpec]] = defaultdict(list)


def _reset_udfs(reset_metric_udfs: bool = True) -> None:
    if reset_metric_udfs:
        _reset_metric_udfs()

    global _multicolumn_udfs, _resolver_specs
    _multicolumn_udfs = defaultdict(list)
    _resolver_specs = defaultdict(list)


def register_dataset_udf(
    col_names: List[str],
    udf_name: Optional[str] = None,
    metrics: Optional[List[MetricSpec]] = None,
    namespace: Optional[str] = None,
    schema_name: str = "",
    anti_metrics: Optional[List[Metric]] = None,
) -> Callable[[Any], Any]:
    """
    Decorator to easily configure UDFs for your data set. Decorate your UDF
    functions, then call generate_udf_dataset_schema() to create a UdfSchema
    that includes the UDFs configured by your decorator parameters. The decorated
    function will automatically be a UDF in the UdfSchema.

    Specify udf_name to give the output of the UDF a name. udf_name
    defautls to the name of the decorated function. Note that all lambdas are
    named "lambda", so omitting udf_name on more than one lambda will result
    in name collisions. If you pass a namespace, it will be prepended to the UDF name.
    Specifying schema_name will register the UDF in a particular schema. If omitted,
    it will be registered to the defualt schema.

    If any metrics are passed via the metrics argument, they will be attached
    to the column produced by the UDF via the schema returned by generate_udf_dataset_schema().
    If metrics is None, the UDF output column will get the metrics determined by
    the other resolvers passed to generate_udf_dataset_schema(), or the STANDARD_UDF_RESOLVER
    by default. Any anti_metrics will be excluded from the metrics attached to the UDF output.
    """

    def decorator_register(func):
        global _multicolumn_udfs, _resolver_specs
        name = udf_name or func.__name__
        name = f"{namespace}.{name}" if namespace else name
        _multicolumn_udfs[schema_name].append(UdfSpec(col_names, {name: func}))
        if metrics:
            _resolver_specs[schema_name].append(ResolverSpec(name, None, deepcopy(metrics)))
        if anti_metrics:
            _resolver_specs[schema_name].append(ResolverSpec(name, None, [MetricSpec(m) for m in anti_metrics], True))

        return func

    return decorator_register


def unregister_udf(udf_name: str, namespace: Optional[str] = None, schema_name: str = "") -> None:
    global _multicolumn_udfs, _resolver_specs
    name = f"{namespace}.{udf_name}" if namespace else udf_name
    if schema_name not in _multicolumn_udfs:
        logger.warn(f"Can't unregister UDF {name} from non-existant schema {schema_name}")
        return

    found = False
    for spec in _multicolumn_udfs[schema_name]:
        if name in spec.udfs:
            found = True
            del spec.udfs[name]
    if not found:
        logger.warn(f"UDF {name} could not be found for unregistering")
    _resolver_specs[schema_name] = list(filter(lambda x: x.column_name != name, _resolver_specs[schema_name]))


def register_type_udf(
    col_type: Type,
    udf_name: Optional[str] = None,
    namespace: Optional[str] = None,
    schema_name: str = "",
    type_mapper: Optional[TypeMapper] = None,
) -> Callable[[Any], Any]:
    """
    Decorator to easily configure UDFs for your data set. Decorate your UDF
    functions, then call generate_udf_dataset_schema() to create a UdfSchema
    that includes the UDFs configured by your decorator parameters. The decorated
    function will automatically be a UDF in the UdfSchema.

    The registered function will be applied to any columns of the specified type.
    Specify udf_name to give the output of the UDF a name. udf_name
    defautls to the name of the decorated function. The output column name is the
    UDF name prefixed with the input column name. Note that all lambdas are
    named "lambda", so omitting udf_name on more than one lambda will result
    in name collisions. If you pass a namespace, it will be prepended to the UDF name.
    Specifying schema_name will register the UDF in a particular schema. If omitted,
    it will be registered to the defualt schema.

    """
    try:
        if not issubclass(col_type, DataType):
            type_mapper = type_mapper or StandardTypeMapper()
            col_type = type(type_mapper(col_type))
    except:  # noqa
        raise ValueError("Column type must be a DataType or Python type mapable to one")

    def decorator_register(func):
        global _multicolumn_udfs, _resolver_specs
        name = udf_name or func.__name__
        name = f"{namespace}.{name}" if namespace else name
        _multicolumn_udfs[schema_name].append(UdfSpec(None, {name: func}, col_type))

        return func

    return decorator_register


def generate_udf_specs(
    other_udf_specs: Optional[List[UdfSpec]] = None,
    schema_name: Union[str, List[str]] = "",
    include_default_schema: bool = True,
) -> List[UdfSpec]:
    """
    Generates a list UdfSpecs that implement the UDFs specified by the
    @register_dataset_udf, @register_type_udf, and @register_metric_udf
    decorators. You can provide a list of other_udf_specs to include in
    addition to those UDFs registered via the decorator.

    For example:

    @register_dataset_udf(col_names=["col1"])
    def add5(x):
        return x + 5

    schema = UdfSchema(STANDARD_RESOLVER, udf_specs=generate_udf_specs())
    why.log(data, schema=schema)

    This will attach a UDF to column "col1" that will generate a new column
    named "add5" containing the values in "col1" incremented by 5. Since these
    are appended to the STANDARD_UDF_RESOLVER, the default metrics are also tracked
    for every column.
    """
    specs = list(other_udf_specs) if other_udf_specs else []
    schema_name = schema_name if isinstance(schema_name, list) else [schema_name]
    if include_default_schema and "" not in schema_name:
        schema_name = [""] + schema_name

    for name in schema_name:
        specs += _multicolumn_udfs[name]
    return specs


DEFAULT_UDF_SCHEMA_RESOLVER = NO_FI_RESOLVER


def udf_schema(
    other_udf_specs: Optional[List[UdfSpec]] = None,
    resolvers: Optional[List[ResolverSpec]] = None,
    types: Optional[Dict[str, Any]] = None,
    default_config: Optional[MetricConfig] = None,
    type_mapper: Optional[TypeMapper] = None,
    cache_size: int = 1024,
    schema_based_automerge: bool = False,
    segments: Optional[Dict[str, SegmentationPartition]] = None,
    validators: Optional[Dict[str, List[Validator]]] = None,
    schema_name: Union[str, List[str]] = "",
    include_default_schema: bool = True,
) -> UdfSchema:
    """
    Returns a UdfSchema that implements any registered UDFs, along with any
    other_udf_specs or resolvers passed in.
    """
    resolver_specs = list(resolvers if resolvers is not None else DEFAULT_UDF_SCHEMA_RESOLVER)
    schema_names = schema_name if isinstance(schema_name, list) else [schema_name]
    if include_default_schema and "" not in schema_names:
        schema_names = [""] + schema_names

    for name in schema_names:
        resolver_specs += _resolver_specs[name]
        validators = generate_validators(validators, name, include_default_schema=True)

    resolver_specs += generate_udf_resolvers(schema_name, include_default_schema)
    return UdfSchema(
        resolver_specs,
        types,
        default_config,
        type_mapper,
        cache_size,
        schema_based_automerge,
        segments,
        validators,
        generate_udf_specs(other_udf_specs, schema_name, include_default_schema),
    )



import logging
import time
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from threading import Timer
from typing import Callable, Type

from dateutil import tz


def current_time_ms() -> int:
    return time.time_ns() // 1_000_000


class TimeGranularity(Enum):
    Second = "Second"
    Minute = "Minute"
    Hour = "Hour"
    Day = "Day"
    Month = "Month"
    Year = "Year"


def truncate_time_ms(t: int, granularity: TimeGranularity) -> int:
    dt = datetime.fromtimestamp(t / 1000, tz=tz.tzutc()).replace(second=0, microsecond=0)

    if granularity == TimeGranularity.Minute:
        trunc = dt
    elif granularity == TimeGranularity.Hour:
        trunc = dt.replace(minute=0)
    elif granularity == TimeGranularity.Day:
        trunc = dt.replace(minute=0, hour=0)
    elif granularity == TimeGranularity.Month:
        trunc = dt.replace(minute=0, hour=0, day=1)
    elif granularity == TimeGranularity.Year:
        trunc = dt.replace(minute=0, hour=0, day=1, month=1)
    else:
        raise ValueError(f"Unsupported granularity: {granularity}")

    return int(trunc.timestamp() * 1000)


@dataclass
class Schedule:
    cadence: TimeGranularity
    interval: int


class FunctionTimer:
    """
    A timer that executes a function repeatedly given a Schedule. It will execute at the bottom of
    that time period. For example, a schedule of Schedule(TimeGranularity.Hour, 1) will execute at
    the start of each hour. If you start the timer 5 minutes before the next hour then it will first
    execute in five minutes, and then each hour after that.
    """

    def __init__(self, schedule: Schedule, fn: Callable, timer_class: Type = Timer) -> None:
        self._logger = logging.getLogger(f"{type(self).__name__}_{id(self)}")
        self._fn = fn
        self._schedule = schedule
        self._running = True
        self._timer_class = timer_class
        now = datetime.now(tz=tz.tzutc())

        # Convert each of the intervals into seconds
        if schedule.cadence == TimeGranularity.Second:
            self.repeat_interval = schedule.interval
            next_second = (now + timedelta(seconds=self._schedule.interval)).replace(microsecond=0)
            initial_interval = min(0, (next_second - now).seconds)
        elif schedule.cadence == TimeGranularity.Minute:
            self.repeat_interval = schedule.interval * 60
            next_minute = (now + timedelta(minutes=self._schedule.interval)).replace(microsecond=0, second=0)
            initial_interval = (next_minute - now).seconds
        elif schedule.cadence == TimeGranularity.Hour:
            self.repeat_interval = schedule.interval * 60 * 60
            next_hour = (now + timedelta(hours=self._schedule.interval)).replace(microsecond=0, second=0, minute=0)
            initial_interval = (next_hour - now).seconds
        elif schedule.cadence == TimeGranularity.Day:
            self.repeat_interval = schedule.interval * 60 * 60 * 24
            next_day = (now + timedelta(days=self._schedule.interval)).replace(
                microsecond=0, second=0, minute=0, hour=0
            )
            initial_interval = (next_day - now).seconds
        elif schedule.cadence == TimeGranularity.Month:
            raise Exception("Can't use Monthly schedule.")

        self._logger.debug(f"scheduled for {initial_interval} seconds from now")
        self._timer = timer_class(initial_interval, self._run)
        self._timer.start()

    def _run(self) -> None:
        self._logger.debug(f"scheduled for {self.repeat_interval} seconds from now")
        self._timer = self._timer_class(self.repeat_interval, self._run)
        self._timer.start()
        self._fn()

    def is_alive(self) -> bool:
        return self._timer.is_alive()

    def stop(self) -> None:
        self._logger.info("Stopping timer")
        if not self._running:
            raise Exception("Timer already stopped")
        self._timer.cancel()
        self._running = False



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.struct_pb2
import sys
import typing
import v0_messages_pb2
import v0_summaries_pb2

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _Op:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _OpEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Op.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    unused: _Op.ValueType  # 0
    LT: _Op.ValueType  # 1
    LE: _Op.ValueType  # 2
    EQ: _Op.ValueType  # 3
    NE: _Op.ValueType  # 4
    GE: _Op.ValueType  # 5
    GT: _Op.ValueType  # 6
    MATCH: _Op.ValueType  # 7
    NOMATCH: _Op.ValueType  # 8
    BTWN: _Op.ValueType  # 9
    IN_SET: _Op.ValueType  # 10
    CONTAIN_SET: _Op.ValueType  # 11
    EQ_SET: _Op.ValueType  # 12
    APPLY_FUNC: _Op.ValueType  # 13
    IN: _Op.ValueType  # 14
    CONTAIN: _Op.ValueType  # 15
    NOT_IN: _Op.ValueType  # 16
    SUM: _Op.ValueType  # 17

class Op(_Op, metaclass=_OpEnumTypeWrapper):
    """constraints specify one of the following binary boolean relationships."""

unused: Op.ValueType  # 0
LT: Op.ValueType  # 1
LE: Op.ValueType  # 2
EQ: Op.ValueType  # 3
NE: Op.ValueType  # 4
GE: Op.ValueType  # 5
GT: Op.ValueType  # 6
MATCH: Op.ValueType  # 7
NOMATCH: Op.ValueType  # 8
BTWN: Op.ValueType  # 9
IN_SET: Op.ValueType  # 10
CONTAIN_SET: Op.ValueType  # 11
EQ_SET: Op.ValueType  # 12
APPLY_FUNC: Op.ValueType  # 13
IN: Op.ValueType  # 14
CONTAIN: Op.ValueType  # 15
NOT_IN: Op.ValueType  # 16
SUM: Op.ValueType  # 17
global___Op = Op

@typing_extensions.final
class SummaryConstraintMsg(google.protobuf.message.Message):
    """Summary constraints specify a relationship between a summary field and a literal value,
    or between two summary fields.
    e.g.     'min' < 6
    'std_dev' < 2.17
    'min' > 'avg'
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    FIRST_FIELD_FIELD_NUMBER: builtins.int
    SECOND_FIELD_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    BETWEEN_FIELD_NUMBER: builtins.int
    REFERENCE_SET_FIELD_NUMBER: builtins.int
    VALUE_STR_FIELD_NUMBER: builtins.int
    OP_FIELD_NUMBER: builtins.int
    VERBOSE_FIELD_NUMBER: builtins.int
    QUANTILE_VALUE_FIELD_NUMBER: builtins.int
    CONTINUOUS_DISTRIBUTION_FIELD_NUMBER: builtins.int
    DISCRETE_DISTRIBUTION_FIELD_NUMBER: builtins.int
    name: builtins.str
    first_field: builtins.str
    second_field: builtins.str
    value: builtins.float
    @property
    def between(self) -> global___SummaryBetweenConstraintMsg: ...
    @property
    def reference_set(self) -> google.protobuf.struct_pb2.ListValue: ...
    value_str: builtins.str
    op: global___Op.ValueType
    verbose: builtins.bool
    quantile_value: builtins.float
    @property
    def continuous_distribution(self) -> global___ReferenceDistributionContinuousMessage: ...
    @property
    def discrete_distribution(self) -> global___ReferenceDistributionDiscreteMessage: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        first_field: builtins.str = ...,
        second_field: builtins.str = ...,
        value: builtins.float = ...,
        between: global___SummaryBetweenConstraintMsg | None = ...,
        reference_set: google.protobuf.struct_pb2.ListValue | None = ...,
        value_str: builtins.str = ...,
        op: global___Op.ValueType = ...,
        verbose: builtins.bool = ...,
        quantile_value: builtins.float = ...,
        continuous_distribution: global___ReferenceDistributionContinuousMessage | None = ...,
        discrete_distribution: global___ReferenceDistributionDiscreteMessage | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["between", b"between", "continuous_distribution", b"continuous_distribution", "discrete_distribution", b"discrete_distribution", "reference_distribution", b"reference_distribution", "reference_set", b"reference_set", "second", b"second", "second_field", b"second_field", "value", b"value", "value_str", b"value_str"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["between", b"between", "continuous_distribution", b"continuous_distribution", "discrete_distribution", b"discrete_distribution", "first_field", b"first_field", "name", b"name", "op", b"op", "quantile_value", b"quantile_value", "reference_distribution", b"reference_distribution", "reference_set", b"reference_set", "second", b"second", "second_field", b"second_field", "value", b"value", "value_str", b"value_str", "verbose", b"verbose"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["reference_distribution", b"reference_distribution"]) -> typing_extensions.Literal["continuous_distribution", "discrete_distribution"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["second", b"second"]) -> typing_extensions.Literal["second_field", "value", "between", "reference_set", "value_str"] | None: ...

global___SummaryConstraintMsg = SummaryConstraintMsg

@typing_extensions.final
class ReferenceDistributionContinuousMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SKETCH_FIELD_NUMBER: builtins.int
    @property
    def sketch(self) -> v0_messages_pb2.KllFloatsSketchMessage: ...
    def __init__(
        self,
        *,
        sketch: v0_messages_pb2.KllFloatsSketchMessage | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["sketch", b"sketch"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["sketch", b"sketch"]) -> None: ...

global___ReferenceDistributionContinuousMessage = ReferenceDistributionContinuousMessage

@typing_extensions.final
class ReferenceDistributionDiscreteMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FREQUENT_ITEMS_FIELD_NUMBER: builtins.int
    UNIQUE_COUNT_FIELD_NUMBER: builtins.int
    TOTAL_COUNT_FIELD_NUMBER: builtins.int
    @property
    def frequent_items(self) -> v0_summaries_pb2.FrequentItemsSummary: ...
    @property
    def unique_count(self) -> v0_summaries_pb2.UniqueCountSummary: ...
    total_count: builtins.float
    def __init__(
        self,
        *,
        frequent_items: v0_summaries_pb2.FrequentItemsSummary | None = ...,
        unique_count: v0_summaries_pb2.UniqueCountSummary | None = ...,
        total_count: builtins.float = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["frequent_items", b"frequent_items", "unique_count", b"unique_count"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["frequent_items", b"frequent_items", "total_count", b"total_count", "unique_count", b"unique_count"]) -> None: ...

global___ReferenceDistributionDiscreteMessage = ReferenceDistributionDiscreteMessage

@typing_extensions.final
class SummaryBetweenConstraintMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SECOND_FIELD_FIELD_NUMBER: builtins.int
    LOWER_VALUE_FIELD_NUMBER: builtins.int
    THIRD_FIELD_FIELD_NUMBER: builtins.int
    UPPER_VALUE_FIELD_NUMBER: builtins.int
    second_field: builtins.str
    lower_value: builtins.float
    third_field: builtins.str
    upper_value: builtins.float
    def __init__(
        self,
        *,
        second_field: builtins.str = ...,
        lower_value: builtins.float = ...,
        third_field: builtins.str = ...,
        upper_value: builtins.float = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["lower", b"lower", "lower_value", b"lower_value", "second_field", b"second_field", "third_field", b"third_field", "upper", b"upper", "upper_value", b"upper_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["lower", b"lower", "lower_value", b"lower_value", "second_field", b"second_field", "third_field", b"third_field", "upper", b"upper", "upper_value", b"upper_value"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["lower", b"lower"]) -> typing_extensions.Literal["second_field", "lower_value"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["upper", b"upper"]) -> typing_extensions.Literal["third_field", "upper_value"] | None: ...

global___SummaryBetweenConstraintMsg = SummaryBetweenConstraintMsg

@typing_extensions.final
class ApplyFunctionMsg(google.protobuf.message.Message):
    """ValueConstraints express a binary boolean relationship between an implied numeric value and a literal, or between a string value and a regex pattern.
    These are applied to every incoming value that is processed by whylogs.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FUNCTION_FIELD_NUMBER: builtins.int
    REFERENCE_VALUE_FIELD_NUMBER: builtins.int
    function: builtins.str
    reference_value: builtins.str
    def __init__(
        self,
        *,
        function: builtins.str = ...,
        reference_value: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["function", b"function", "reference_value", b"reference_value"]) -> None: ...

global___ApplyFunctionMsg = ApplyFunctionMsg

@typing_extensions.final
class ValueConstraintMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    REGEX_PATTERN_FIELD_NUMBER: builtins.int
    VALUE_SET_FIELD_NUMBER: builtins.int
    FUNCTION_FIELD_NUMBER: builtins.int
    OP_FIELD_NUMBER: builtins.int
    VERBOSE_FIELD_NUMBER: builtins.int
    TOTAL_FIELD_NUMBER: builtins.int
    FAILURES_FIELD_NUMBER: builtins.int
    name: builtins.str
    value: builtins.float
    regex_pattern: builtins.str
    @property
    def value_set(self) -> google.protobuf.struct_pb2.ListValue: ...
    @property
    def function(self) -> global___ApplyFunctionMsg: ...
    op: global___Op.ValueType
    verbose: builtins.bool
    total: builtins.int
    failures: builtins.int
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        value: builtins.float = ...,
        regex_pattern: builtins.str = ...,
        value_set: google.protobuf.struct_pb2.ListValue | None = ...,
        function: global___ApplyFunctionMsg | None = ...,
        op: global___Op.ValueType = ...,
        verbose: builtins.bool = ...,
        total: builtins.int = ...,
        failures: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["function", b"function", "regex_pattern", b"regex_pattern", "second_field", b"second_field", "value", b"value", "value_set", b"value_set"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["failures", b"failures", "function", b"function", "name", b"name", "op", b"op", "regex_pattern", b"regex_pattern", "second_field", b"second_field", "total", b"total", "value", b"value", "value_set", b"value_set", "verbose", b"verbose"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["second_field", b"second_field"]) -> typing_extensions.Literal["value", "regex_pattern", "value_set", "function"] | None: ...

global___ValueConstraintMsg = ValueConstraintMsg

@typing_extensions.final
class MultiColumnValueConstraintMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    DEPENDENT_COLUMNS_FIELD_NUMBER: builtins.int
    DEPENDENT_COLUMN_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    VALUE_SET_FIELD_NUMBER: builtins.int
    REFERENCE_COLUMNS_FIELD_NUMBER: builtins.int
    OP_FIELD_NUMBER: builtins.int
    VERBOSE_FIELD_NUMBER: builtins.int
    INTERNAL_DEPENDENT_COLUMNS_OP_FIELD_NUMBER: builtins.int
    TOTAL_FIELD_NUMBER: builtins.int
    FAILURES_FIELD_NUMBER: builtins.int
    name: builtins.str
    @property
    def dependent_columns(self) -> google.protobuf.struct_pb2.ListValue: ...
    dependent_column: builtins.str
    value: builtins.float
    @property
    def value_set(self) -> google.protobuf.struct_pb2.ListValue: ...
    @property
    def reference_columns(self) -> google.protobuf.struct_pb2.ListValue: ...
    op: global___Op.ValueType
    verbose: builtins.bool
    internal_dependent_columns_op: global___Op.ValueType
    total: builtins.int
    failures: builtins.int
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        dependent_columns: google.protobuf.struct_pb2.ListValue | None = ...,
        dependent_column: builtins.str = ...,
        value: builtins.float = ...,
        value_set: google.protobuf.struct_pb2.ListValue | None = ...,
        reference_columns: google.protobuf.struct_pb2.ListValue | None = ...,
        op: global___Op.ValueType = ...,
        verbose: builtins.bool = ...,
        internal_dependent_columns_op: global___Op.ValueType = ...,
        total: builtins.int = ...,
        failures: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["dependent", b"dependent", "dependent_column", b"dependent_column", "dependent_columns", b"dependent_columns", "reference", b"reference", "reference_columns", b"reference_columns", "value", b"value", "value_set", b"value_set"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["dependent", b"dependent", "dependent_column", b"dependent_column", "dependent_columns", b"dependent_columns", "failures", b"failures", "internal_dependent_columns_op", b"internal_dependent_columns_op", "name", b"name", "op", b"op", "reference", b"reference", "reference_columns", b"reference_columns", "total", b"total", "value", b"value", "value_set", b"value_set", "verbose", b"verbose"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["dependent", b"dependent"]) -> typing_extensions.Literal["dependent_columns", "dependent_column"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["reference", b"reference"]) -> typing_extensions.Literal["value", "value_set", "reference_columns"] | None: ...

global___MultiColumnValueConstraintMsg = MultiColumnValueConstraintMsg

@typing_extensions.final
class ValueConstraintMsgs(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONSTRAINTS_FIELD_NUMBER: builtins.int
    MULTI_COLUMN_CONSTRAINTS_FIELD_NUMBER: builtins.int
    @property
    def constraints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ValueConstraintMsg]: ...
    @property
    def multi_column_constraints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MultiColumnValueConstraintMsg]: ...
    def __init__(
        self,
        *,
        constraints: collections.abc.Iterable[global___ValueConstraintMsg] | None = ...,
        multi_column_constraints: collections.abc.Iterable[global___MultiColumnValueConstraintMsg] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["constraints", b"constraints", "multi_column_constraints", b"multi_column_constraints"]) -> None: ...

global___ValueConstraintMsgs = ValueConstraintMsgs

@typing_extensions.final
class SummaryConstraintMsgs(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONSTRAINTS_FIELD_NUMBER: builtins.int
    @property
    def constraints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SummaryConstraintMsg]: ...
    def __init__(
        self,
        *,
        constraints: collections.abc.Iterable[global___SummaryConstraintMsg] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["constraints", b"constraints"]) -> None: ...

global___SummaryConstraintMsgs = SummaryConstraintMsgs

@typing_extensions.final
class DatasetConstraintMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class ValueConstraintsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___ValueConstraintMsgs: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___ValueConstraintMsgs | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing_extensions.final
    class SummaryConstraintsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___SummaryConstraintMsgs: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___SummaryConstraintMsgs | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    PROPERTIES_FIELD_NUMBER: builtins.int
    VALUE_CONSTRAINTS_FIELD_NUMBER: builtins.int
    SUMMARY_CONSTRAINTS_FIELD_NUMBER: builtins.int
    TABLE_SHAPE_CONSTRAINTS_FIELD_NUMBER: builtins.int
    MULTI_COLUMN_VALUE_CONSTRAINTS_FIELD_NUMBER: builtins.int
    @property
    def properties(self) -> v0_messages_pb2.DatasetPropertiesV0: ...
    @property
    def value_constraints(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___ValueConstraintMsgs]: ...
    @property
    def summary_constraints(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___SummaryConstraintMsgs]: ...
    @property
    def table_shape_constraints(self) -> global___SummaryConstraintMsgs: ...
    @property
    def multi_column_value_constraints(self) -> global___ValueConstraintMsgs: ...
    def __init__(
        self,
        *,
        properties: v0_messages_pb2.DatasetPropertiesV0 | None = ...,
        value_constraints: collections.abc.Mapping[builtins.str, global___ValueConstraintMsgs] | None = ...,
        summary_constraints: collections.abc.Mapping[builtins.str, global___SummaryConstraintMsgs] | None = ...,
        table_shape_constraints: global___SummaryConstraintMsgs | None = ...,
        multi_column_value_constraints: global___ValueConstraintMsgs | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["multi_column_value_constraints", b"multi_column_value_constraints", "properties", b"properties", "table_shape_constraints", b"table_shape_constraints"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["multi_column_value_constraints", b"multi_column_value_constraints", "properties", b"properties", "summary_constraints", b"summary_constraints", "table_shape_constraints", b"table_shape_constraints", "value_constraints", b"value_constraints"]) -> None: ...

global___DatasetConstraintMsg = DatasetConstraintMsg



from abc import abstractmethod
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

from whylogs.core.stubs import pd

Row = Dict[str, Any]
TrackData = Union[pd.DataFrame, Row, List[Row]]

StatusType = TypeVar("StatusType")


class DataLogger(Generic[StatusType]):
    """
    Interface for data loggers. This type has all of the methods that people would need for actually
    logging data. Its potentially useful to use this as a return type instead of the process/thread logger
    since those have a lot more private methods that aren't useful to the end user.
    """

    @abstractmethod
    def status(self, timeout: Optional[float] = 1.0) -> StatusType:
        """
        Get diagnostic information about the state of the data logger
        """
        raise NotImplementedError()

    @abstractmethod
    def log(
        self,
        data: TrackData,
        timestamp_ms: Optional[int] = None,  # The timestamp that the data happened at
        sync: bool = False,
    ) -> None:
        """
        Log some data.

        Parameters:
            data: The data to log. This can either be a pandas data frame, a row (dictionary of str to str/int/float/etc),
                or a list of rows.
            timestamp_ms: The timestamp of the data. If this isn't supplied then it is assumed to have happened now.
            sync: Whether or not to perform this action synchronously. By default, this is an asynchronous operation.
                You can make this synchronous in order to react to errors. Mostly useful when initially setting up
                logging since the only errors that can be responded to are data format related.
        """
        raise NotImplementedError()

    @abstractmethod
    def flush(self) -> None:
        """
        Flush the internal state, causing everything to be written using the configured writers.
        """
        raise NotImplementedError()

    @abstractmethod
    def close(self) -> None:
        """
        Close the logger, causing all resources to be released.
        """
        raise NotImplementedError()



import logging
from dataclasses import dataclass, field
from enum import Enum
from itertools import chain
from typing import List, Optional

from whylogs.core.metrics import StandardMetric
from whylogs.core.metrics.metrics import MetricConfig, OperationResult, register_metric
from whylogs.core.metrics.multimetric import MultiMetric
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.proto import MetricMessage
from whylogs.core.stubs import np, sklp
from whylogs.experimental.extras.matrix_component import MatrixComponent

logger = logging.getLogger(__name__)


class DistanceFunction(Enum):
    euclidean = sklp.euclidean_distances
    cosine = sklp.cosine_distances


@dataclass(frozen=True)
class EmbeddingConfig(MetricConfig):
    """
    The rows of references are the reference vectors. A shape of (1, 1) indicates
    there's no reference matrix and the metric will not be updatable. It should still
    be mergeable with compatible metrics (details of compatability TBD). The rows
    must be in the same order as labels. If labels are not provided, "0" ... "n" will
    be used, where n is the number of rows in the references matrix.
    """

    references: np.ndarray = field(default_factory=lambda: np.zeros((1, 1)))
    labels: Optional[List[str]] = None
    distance_fn: DistanceFunction = DistanceFunction.cosine
    serialize_references: bool = True  # should references be included in protobuf message?

    # TODO: limit refeence size

    def __post_init__(self) -> None:
        if len(self.references.shape) != 2:
            raise ValueError("Embedding reference matrix must be 2 dimensional")

        if self.labels:
            if len(self.labels) != self.references.shape[0]:
                raise ValueError(
                    f"Number of labels ({len(self.labels)}) must match number of reference vectors ({self.references.shape[1]})"
                )


@dataclass
class EmbeddingMetric(MultiMetric):
    references: MatrixComponent
    labels: List[str]
    distance_fn: DistanceFunction
    serialize_references: bool

    def __post_init__(self):
        """
        Each label has a {label}_distance submetric that tracks the distribution of
        distances from the label's reference vector to every logged vector.

        There is also a closest submetric that tracks how often a label's reference
        vector is the closest to the logged vectors.
        """

        submetrics = {
            f"{label}_distance": {
                "distribution": StandardMetric.distribution.zero(),
                "counts": StandardMetric.counts.zero(),
                "types": StandardMetric.types.zero(),
                "cardinality": StandardMetric.cardinality.zero(),
            }
            for label in self.labels
        }
        submetrics.update(
            {
                "closest": {
                    "frequent_items": StandardMetric.frequent_items.zero(),
                    "counts": StandardMetric.counts.zero(),
                    "types": StandardMetric.types.zero(),
                    "cardinality": StandardMetric.cardinality.zero(),
                }
            }
        )
        super().__init__(submetrics)
        # sort labels and permute reference matrix rows to match to support deserialization
        label_indices = np.argsort(self.labels).tolist()
        self.labels = sorted(self.labels)
        self.references = MatrixComponent(self.references.value[label_indices, :])

    @property
    def namespace(self) -> str:
        return "embedding"

    def merge(self, other: "EmbeddingMetric") -> "EmbeddingMetric":
        if self.references.value.shape != other.references.value.shape:
            if other.references.value.shape == (1, 1):
                # TODO: handle merging with other.serialize_references==False better
                # The (1, 1) shape indicates the other metric was created without a reference matrix.
                # It can't have meaningful data in it, so just return myself
                logger.warning("Attempt to merge with unconfigured EmbeddingMetric; ignored")
                return self
            if self.references.value.shape == (1, 1):
                # See comment above
                logger.warning("Attempt to merge with unconfigured EmbeddingMetric; ignored")
                return other
            raise ValueError("Attempt to merge incompatible EbeddingMetrics")

        if (
            self.labels != other.labels
            or not (self.references.value == other.references.value).all()
            # TODO: maybe   or self.distance_fn != other.distance_fn ? warn if != ?
        ):
            raise ValueError("Attempt to merge incompatible EbeddingMetrics")

        result = EmbeddingMetric(self.references, self.labels, self.distance_fn, self.serialize_references)
        result.submetrics = self.merge_submetrics(other)
        return result

    def to_protobuf(self) -> MetricMessage:
        msg = {}
        for sub_name, metrics in self.submetrics.items():
            for namespace, metric in metrics.items():
                sub_msg = metric.to_protobuf()
                for comp_name, comp_msg in sub_msg.metric_components.items():
                    msg[f"{sub_name}:{namespace}/{comp_name}"] = comp_msg
        if self.serialize_references:
            msg["references"] = self.references.to_protobuf()

        return MetricMessage(metric_components=msg)

    def _update_submetrics(self, submetric: str, data: PreprocessedColumn) -> None:
        for key in self.submetrics[submetric].keys():
            self.submetrics[submetric][key].columnar_update(data)

    def columnar_update(self, data: PreprocessedColumn) -> OperationResult:
        reference_dim = self.references.value.shape[1]  # number of columns in reference matrix
        successes = 0
        failures = 0
        pandas_tensors = data.pandas.tensors if data.pandas.tensors is not None else []
        for matrix in chain(data.list.tensors or [], pandas_tensors):  # TODO: stack these
            if len(matrix.shape) == 1:
                matrix = matrix.reshape((1, matrix.shape[0]))
            if len(matrix.shape) != 2 or matrix.shape[1] != reference_dim:
                logger.warn(
                    f"EmbeddingMetric requires 1 x {reference_dim} matrices; got tensor with shape {matrix.shape}"
                )
                failures += 1
                continue

            ref_dists = self.distance_fn(matrix, self.references.value)  # type: ignore
            ref_closest = np.argmin(ref_dists, axis=1)

            for i in range(ref_dists.shape[1]):
                self._update_submetrics(f"{self.labels[i]}_distance", PreprocessedColumn.apply(ref_dists[:, i]))

            closest = [self.labels[i] for i in ref_closest]
            self._update_submetrics("closest", PreprocessedColumn.apply(np.asarray(closest)))
            successes += 1

        return OperationResult(failures, successes)

    @classmethod
    def from_protobuf(cls, msg: MetricMessage) -> "EmbeddingMetric":
        if "references" in msg.metric_components:
            references = MatrixComponent.from_protobuf(msg.metric_components["references"])
            msg.metric_components.pop("references")  # it's not a submetric's component
            serialize_references = True
        else:
            references = np.zeros((1, 1))  # indicate I don't have a usuable reference matrix
            serialize_references = False

        submetrics = EmbeddingMetric.submetrics_from_protobuf(msg)

        # figure out what my labels were from the {label}_distance submetric names
        labels: List[str] = []
        for submetric_name in submetrics.keys():
            if submetric_name.endswith("_distance"):
                labels.append(submetric_name[:-9])
        labels = sorted(labels)  # the rows should already be in this order

        result = EmbeddingMetric(
            references=references,
            labels=labels,
            distance_fn=DistanceFunction.cosine,  # not updatable after deserialization
            serialize_references=serialize_references,
        )
        result.submetrics = submetrics
        return result

    @classmethod
    def zero(cls, cfg: Optional[EmbeddingConfig] = None) -> "EmbeddingMetric":
        cfg = cfg or EmbeddingConfig()
        if not isinstance(cfg, EmbeddingConfig):
            raise ValueError("EmbeddingMetric.zero() requires EmbeddingConfig argument")

        return EmbeddingMetric(
            references=MatrixComponent(cfg.references),
            labels=cfg.labels or [str(i) for i in range(cfg.references.shape[0])],
            distance_fn=cfg.distance_fn,
            serialize_references=cfg.serialize_references,
        )


# Register it so Multimetric and ProfileView can deserialize
register_metric(EmbeddingMetric)



from typing import List

from whylogs.core import CardinalityThresholds
from whylogs.core.constraints.factories import frequent_strings_in_reference_set
from whylogs.core.constraints.metric_constraints import MetricConstraint
from whylogs.core.utils import get_cardinality_estimate
from whylogs.core.view.column_profile_view import ColumnProfileView


def generate_column_frequent_items_constraints(
    column_name: str, column_profile: ColumnProfileView
) -> List[MetricConstraint]:
    constraints = []
    fi_list = [fi.value for fi in column_profile.get_metric("frequent_items").to_summary_dict()["frequent_strings"]]
    cardinality_estimate = get_cardinality_estimate(column_profile)
    if (
        0 < cardinality_estimate.get("est") <= CardinalityThresholds.few
        and cardinality_estimate.get("unique_pct") < CardinalityThresholds.proportionately_few
    ):
        fi_set = set(fi_list)
        constraints.append(frequent_strings_in_reference_set(column_name, fi_set))
    return constraints



import logging
from dataclasses import dataclass
from typing import Dict, Optional

from whylogs.core import ColumnProfileView, DatasetProfileView
from whylogs.core.metrics import Metric
from whylogs.core.metrics.column_metrics import ColumnCountsMetric
from whylogs.core.metrics.compound_metric import CompoundMetric
from whylogs.core.metrics.condition_count_metric import ConditionCountMetric
from whylogs.core.metrics.metric_components import IntegralComponent
from whylogs.core.metrics.multimetric import MultiMetric

try:
    from PIL.Image import Image as ImageType  # noqa trigger ImportError if unavailable

    from whylogs.extras.image_metric import ImageMetric
except ImportError:
    # dummy ImageMetric class so nothing will be an instance of it
    class ImageMetric:  # type: ignore[no-redef]
        pass


logger = logging.getLogger(__name__)


@dataclass
class FeatureFlags:
    uncompound_condition_count: bool = False


def _uncompound_metric_feature_flag() -> bool:
    """
    v0 whylabs doesn't understand compound metrics. If this is True, turn
    each submetric in a compound metric or MultiMetric into its own column
    in the profile so that v0 whylabs will only see metrics it understands.
    """
    return True


def _uncompound_condition_count_feature_flag(flags: Optional[FeatureFlags] = None) -> bool:
    """
    v0 whylabs doesn't understand condition count metrics. If this is True, turn
    each condition into its three ColumnCountMetric columns in the profile
    so that v0 whylabs will only see metrics it understands.
    """
    if flags is None:
        return False
    return flags.uncompound_condition_count


def _uncompound_performance_estimation_magic_string() -> str:
    """
    Column name prefix for performance estimation metrics
    """
    return "Ω.whylabs.output_performance_estimation."


def _uncompound_performance_estimation_feature_flag() -> bool:
    return True


def _v0_compatible_image_feature_flag() -> bool:
    """
    v0 whylogs only supported logging a single image in a profile.
    v1 supports multiple image in a profile by giving each image "channel"
    a unique name. If this returns True, use the old V0 column naming
    convention that only supports a single image.
    """
    return False


def _condition_count_magic_string() -> str:
    """
    Column name prefix for uncompounded ConditionCountMetric columns
    """
    return "Ω.whylabs.condition."


def _uncompounded_column_name(column_name: str, metric_name: str, submetric_name: str, metric: Metric) -> str:
    if isinstance(metric, ImageMetric) and _v0_compatible_image_feature_flag():
        return submetric_name
    return f"{column_name}.{metric_name}.{submetric_name}"


def _uncompound_metric(col_name: str, metric_name: str, metric: CompoundMetric) -> Dict[str, ColumnProfileView]:
    if not _uncompound_metric_feature_flag():
        return dict()

    result: Dict[str, ColumnProfileView] = dict()
    for submetric_name, submetric in metric.submetrics.items():
        new_col_name = _uncompounded_column_name(col_name, metric_name, submetric_name, metric)
        result[new_col_name] = ColumnProfileView({submetric.namespace: submetric})
    return result


def _uncompounded_multimetric_column_name(column_name: str, submetric_name: str, metric: Metric) -> str:
    if isinstance(metric, ImageMetric) and _v0_compatible_image_feature_flag():
        return submetric_name
    return f"{column_name}.{submetric_name}"


def _uncompound_multimetric(col_name: str, metric_name: str, metric: MultiMetric) -> Dict[str, ColumnProfileView]:
    if not _uncompound_metric_feature_flag():
        return dict()

    result: Dict[str, ColumnProfileView] = dict()
    for submetric_name, submetrics in metric.submetrics.items():
        new_col_name = _uncompounded_multimetric_column_name(col_name, submetric_name, metric)
        result[new_col_name] = ColumnProfileView(submetrics)

    return result


def _uncompound_condition_count(
    col_name: str, metric_name: str, metric: ConditionCountMetric, flags: Optional[FeatureFlags] = None
) -> Dict[str, ColumnProfileView]:
    if not _uncompound_condition_count_feature_flag(flags):
        return dict()

    result: Dict[str, ColumnProfileView] = dict()
    for condition_name, count_component in metric.matches.items():
        new_col_name = f"{_condition_count_magic_string()}{col_name}.{condition_name}.total"
        new_metric = ColumnCountsMetric(
            n=metric.total,  # total condition evaluations
            null=IntegralComponent(0),  # unused
            nan=IntegralComponent(0),  # unused
            inf=IntegralComponent(0),  # unused
        )
        result[new_col_name] = ColumnProfileView({ColumnCountsMetric.get_namespace(): new_metric})
        new_col_name = f"{_condition_count_magic_string()}{col_name}.{condition_name}.matches"
        new_metric = ColumnCountsMetric(
            n=count_component,  # count of evaluations that matched condition
            null=IntegralComponent(0),  # unused
            nan=IntegralComponent(0),  # unused
            inf=IntegralComponent(0),  # unused
        )
        result[new_col_name] = ColumnProfileView({ColumnCountsMetric.get_namespace(): new_metric})
        new_col_name = f"{_condition_count_magic_string()}{col_name}.{condition_name}.non_matches"
        new_metric = ColumnCountsMetric(
            n=IntegralComponent(metric.total.value - count_component.value),  # evaluations that didn't match
            null=IntegralComponent(0),  # unused
            nan=IntegralComponent(0),  # unused
            inf=IntegralComponent(0),  # unused
        )
        result[new_col_name] = ColumnProfileView({ColumnCountsMetric.get_namespace(): new_metric})

    return result


def _uncompound_dataset_profile(prof: DatasetProfileView, flags: Optional[FeatureFlags] = None) -> DatasetProfileView:
    """
    v0 whylabs doesn't understand compound metrics. This creates a new column for
    each submetric in a compound metric so that whylabs only sees metrics it understands.
    """
    new_prof = DatasetProfileView(
        columns=prof._columns,
        dataset_timestamp=prof._dataset_timestamp,
        creation_timestamp=prof._creation_timestamp,
        metrics=prof._metrics,
        metadata=prof._metadata,
    )
    new_columns: Dict[str, ColumnProfileView] = dict()
    for col_name, col_prof in new_prof._columns.items():
        for metric_name, metric in col_prof._metrics.items():
            if isinstance(metric, CompoundMetric):
                new_columns.update(_uncompound_metric(col_name, metric_name, metric))
            if isinstance(metric, MultiMetric):
                new_columns.update(_uncompound_multimetric(col_name, metric_name, metric))
            if isinstance(metric, ConditionCountMetric):
                new_columns.update(_uncompound_condition_count(col_name, metric_name, metric, flags))

    new_prof._columns.update(new_columns)
    return new_prof



import logging
from typing import Optional

from whylogs.api.whylabs.session.config import _INIT_DOCS, InitConfig, SessionConfig
from whylogs.api.whylabs.session.session import (
    ApiKeySession,
    GuestSession,
    LocalSession,
    Session,
)
from whylogs.api.whylabs.session.session_types import InteractiveLogger as il
from whylogs.api.whylabs.session.session_types import SessionType

logger = logging.getLogger(__name__)


class SessionManager:
    __instance: Optional["SessionManager"] = None

    def __init__(
        self,
        config: SessionConfig,
    ):
        self.session: Session
        session_type = config.get_session_type()
        if session_type == SessionType.LOCAL:
            self.session = LocalSession(config)
        elif session_type == SessionType.WHYLABS_ANONYMOUS:
            self.session = GuestSession(config)
        elif session_type == SessionType.WHYLABS:
            self.session = ApiKeySession(config)
        else:
            raise ValueError(f"Unknown session type: {session_type}")

    @staticmethod
    def init(session_config: SessionConfig) -> "SessionManager":
        if SessionManager.__instance is None:
            SessionManager.__instance = SessionManager(session_config)
        else:
            logger.debug("SessionManager is already initialized. Ignoring call to init()")

        return SessionManager.__instance

    @staticmethod
    def reset() -> None:
        SessionManager.__instance = None

    @staticmethod
    def get_instance() -> Optional["SessionManager"]:
        return SessionManager.__instance

    @staticmethod
    def is_active() -> bool:
        return SessionManager.get_instance() is not None


def init(  # type: ignore
    reinit: bool = False,
    allow_anonymous: bool = True,
    allow_local: bool = False,
    whylabs_api_key: Optional[str] = None,
    default_dataset_id: Optional[str] = None,
    config_path: Optional[str] = None,
    **kwargs,
) -> Session:
    """
    Set up authentication for this whylogs logging session. There are three modes that you can authentiate in.

    1. WHYLABS: Data is sent to WhyLabs and is associated with a specific WhyLabs account. You can get a WhyLabs api
        key from the WhyLabs Settings page after logging in.
    2. WHYLABS_ANONYMOUS: Data is sent to WhyLabs, but no authentication happens and no WhyLabs account is required.
        Sessions can be claimed into an account later on the WhyLabs website.
    3. LOCAL: No authentication. No data is automatically sent anywhere. Use this if you want to explore profiles
        locally or manually upload them somewhere.

    Typically, you should only have to put `why.init()` with no arguments at the start of your application/notebook/script.
    The arguments allow for some customization of the logic that determines the session type. Here is the priority order:

    - If there is an api key directly supplied to init, then use it and authenticate session as WHYLABS.
    - If there is an api key in the environment variable WHYLABS_API_KEY, then use it and authenticate session as WHYLABS.
    - If there is an api key in the whylogs config file, then use it and authenticate session as WHYLABS.
    - If we're in an interractive environment (notebook, colab, etc.) then prompt the user to pick a method explicitly.
        The options are determined by the allow* argument values to init().
    - If allow_anonymous is True, then authenticate session as WHYLABS_ANONYMOUS.
    - If allow_local is True, then authenticate session as LOCAL.

    Args:
        session_type: Deprecated, use allow_anonymous and allow_local instead
        reinit: Normally, init() is idempotent, so you can run it over and over again in a notebook without any issues, for example.
            If reinit=True then it will run the initialization logic again, so you can switch authentication methods without restarting.
        allow_anonymous: If True, then the user will be able to choose WHYLABS_ANONYMOUS if no other authentication method is found.
        allow_local: If True, then the user will be able to choose LOCAL if no other authentication method is found.
        whylabs_api_key: A WhyLabs api key to use for uploading profiles. There are other ways that you can set an api key that don't
            require direclty embedding it in code, like setting WHYLABS_API_KEY env variable or supplying the api key interractively
            via the init() prompt in a notebook.
        default_dataset_id: The default dataset id to use for uploading profiles. This is only used if the session is authenticated.
            This is a convenience argument so that you don't have to supply the dataset id every time you upload a profile if
            you're only using a single dataset id.

    """
    if reinit:
        SessionManager.reset()

    manager: SessionManager = SessionManager._SessionManager__instance  # type: ignore
    if manager is not None:
        return manager.session

    session_config = SessionConfig(
        InitConfig(
            allow_anonymous=allow_anonymous,
            allow_local=allow_local,
            whylabs_api_key=whylabs_api_key,
            default_dataset_id=default_dataset_id,
            config_path=config_path,
            force_local=kwargs.get("force_local", False),
        )
    )

    try:
        manager = SessionManager.init(session_config)
        session_config.notify_session_type()
        return manager.session
    except PermissionError as e:
        # TODO PR this implies that we need disk access to work correctly, but isn't that already the case
        # because we write profilfes to disk as tmp files?
        logger.warning("Could not create or read configuration file for session. Profiles won't be uploaded.", e)
        raise e
    except Exception as e:
        logger.warning("Could not initialize session", e)
        raise e


def get_current_session() -> Optional[Session]:
    manager = SessionManager.get_instance()
    if manager is not None:
        return manager.session

    il.warning_once(
        f"No session found. Call whylogs.init() to initialize a session and authenticate. See {_INIT_DOCS} for more information.",
        logger.warning,
    )

    return None


def _default_init() -> Session:
    """
    For internal use. This initializes a default session for the user if they don't call why.init() themselves.
    This will behave as though they called why.init() with no arguments and print out a warning with a link to the docs.
    """
    manager = SessionManager.get_instance()
    if manager is None:
        il.warning_once("Initializing default session because no session was found.", logger.warning)

        # To be safe, don't allow default session to be anonymous if this is happening as a side effect
        # that users don't know about.
        return init(allow_anonymous=False, allow_local=True, force_local=True)
    else:
        return manager.session



import re
from collections import namedtuple
from copy import deepcopy
from dataclasses import dataclass
from logging import getLogger
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from whylogs.core.configs import SummaryConfig
from whylogs.core.metrics.metrics import Metric
from whylogs.core.predicate_parser import _METRIC_REF, _PROFILE_REF, _tokenize
from whylogs.core.utils import deprecated
from whylogs.core.view.column_profile_view import ColumnProfileView
from whylogs.core.view.dataset_profile_view import DatasetProfileView

logger = getLogger(__name__)


ReportResult = namedtuple("ReportResult", "name passed failed summary")


@dataclass
class MetricsSelector:
    metric_name: str
    column_name: Optional[str] = None  # <- TODO: return a collection of columns
    metrics_resolver: Optional[Callable[[DatasetProfileView], List[Metric]]] = None

    def column_profile(self, profile: Optional[DatasetProfileView]) -> Optional[ColumnProfileView]:
        if profile is None:
            return None
        column_profile_view = profile.get_column(self.column_name)
        if self.column_name and column_profile_view is None:
            logger.warning(
                f"No column name {self.column_name} found in profile, available columns are: {profile.get_columns()}"
            )
        return column_profile_view

    def apply(self, profile: DatasetProfileView) -> List[Metric]:
        if self.metrics_resolver is not None:
            custom_result = self.metrics_resolver(profile)
            if isinstance(custom_result, List):
                return custom_result
            else:
                raise ValueError("metrics_resolver must return a list of Metrics if defined!")

        results: List[Metric] = []
        column_profile_view = self.column_profile(profile)
        if column_profile_view is None:
            return results
        metric = column_profile_view.get_metric(self.metric_name)
        if metric is None:
            logger.warning(
                f"{self.metric_name} not found in {self.column_name} available metrics are: "
                f"{column_profile_view.get_metric_component_paths()}"
            )
        else:
            results.append(metric)
        return results


def pretty_display(metric_name: str, summary: Dict[str, Any]) -> Dict[str, Any]:
    pretty_summary: Dict[str, Any] = {"metric": metric_name}

    for key, item in summary.items():
        if key != "frequent_strings":
            pretty_summary[key] = item
        else:
            pretty_summary[f"{key}_top_10"] = [f"{x.value}:{x.est}" for x in item[:10]]
    return pretty_summary


@dataclass
class MetricConstraint:
    condition: Callable[[Metric], bool]
    name: str
    metric_selector: MetricsSelector
    require_column_existence: bool = True

    def _validate_metrics(self, metrics: List[Metric]) -> bool:
        for metric in metrics:
            if not self.condition(metric):
                return False
        return True

    def _get_metric_summary(self, metrics: List[Metric]) -> Optional[Dict[str, Any]]:
        if len(metrics) == 1:  # Only returns a summary for single metrics.
            metric_summary = pretty_display(metrics[0].namespace, metrics[0].to_summary_dict())
            return metric_summary
        return None

    @deprecated(message="Please use validate_profile()")
    def validate(self, dataset_profile: DatasetProfileView) -> bool:
        # custom metric resolver allows empty metrics
        if self.metric_selector is None:
            raise ValueError("can't call validate with an empty metric selector")
        metric_selector: MetricsSelector = self.metric_selector
        if metric_selector.metrics_resolver is not None:
            metrics = metric_selector.apply(dataset_profile)
            logger.info(f"validate using custom metric selector and found {metrics}")
            return self._validate_metrics(metrics)

        # standard selector requires a column be resolved
        column_profile = metric_selector.column_profile(dataset_profile)
        if column_profile is None:
            if self.require_column_existence:
                raise ValueError(
                    f"Could not get column {metric_selector.column_name}" f"from column {dataset_profile.get_columns()}"
                )
            else:
                logger.info(
                    f"validate could not find column {metric_selector.column_name} "
                    "but require_column_existence is false so returning True"
                )
                return True
        metrics = metric_selector.apply(dataset_profile)
        if metrics is None or len(metrics) == 0:
            if self.require_column_existence:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} so returning False. Available metrics on column are: "
                    f"{column_profile.get_metric_component_paths()}"
                )
                return False
            else:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} but require_column_existence is set so returning True. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return True
        return self._validate_metrics(metrics)

    def validate_profile(self, dataset_profile: DatasetProfileView) -> Tuple[bool, Optional[Dict[str, Any]]]:
        # custom metric resolver allows empty metrics
        if self.metric_selector is None:
            raise ValueError("can't call validate with an empty metric selector")
        metric_selector: MetricsSelector = self.metric_selector
        if metric_selector.metrics_resolver is not None:
            metrics = metric_selector.apply(dataset_profile)
            logger.info(f"validate using custom metric selector and found {metrics}")
            validate_result = self._validate_metrics(metrics)
            metric_summary = self._get_metric_summary(metrics)
            return (validate_result, metric_summary)

        # standard selector requires a column be resolved
        column_profile = metric_selector.column_profile(dataset_profile)
        if column_profile is None:
            if self.require_column_existence:
                raise ValueError(
                    f"Could not get column {metric_selector.column_name}" f"from column {dataset_profile.get_columns()}"
                )
            else:
                logger.info(
                    f"validate could not find column {metric_selector.column_name} "
                    "but require_column_existence is false so returning True"
                )
                return (True, None)
        metrics = metric_selector.apply(dataset_profile)
        if metrics is None or len(metrics) == 0:
            if self.require_column_existence:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} so returning False. Available metrics on column are: "
                    f"{column_profile.get_metric_component_paths()}"
                )
                return (False, None)
            else:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} but require_column_existence is not set so returning True. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return (True, None)
        if len(metrics) == 1 and metrics[0].namespace == "distribution":
            if metrics[0].kll.value.is_empty() and self.require_column_existence:
                logger.info(
                    f"validate reached empty distribution metrics from column "
                    f"{metric_selector.column_name} but require_column_existence is set so returning False. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return (False, None)
            if metrics[0].kll.value.is_empty() and not self.require_column_existence:
                logger.info(
                    f"validate reached empty distribution metrics from column "
                    f"{metric_selector.column_name} but require_column_existence is not set so returning True. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return (True, None)

        validate_result = self._validate_metrics(metrics)
        metric_summary = self._get_metric_summary(metrics)
        return (validate_result, metric_summary)


class MissingMetric(Exception):
    """
    DatasetConstraint conditions can raise this exception to indicate that
    a required metric or column could not be found in the DatasetProfileView.
    The DatasetConstraint is responsible for handling this exception.
    """


@dataclass(frozen=True)
class DatasetComparisonConstraint:
    """
    Implements dataset-level constraints that require a reference dataset to
    be compared against. The condition Callable takes the DatasetProfileView
    of the dataset to be validated as well as the DatasetProfileView of the
    reference dataset and returns a boolean indicating whether the condition
    is satisfied as well as a dictionary with summary information about the
    validation.

    """

    condition: Callable[[DatasetProfileView, DatasetProfileView], Tuple[bool, Dict[str, Metric]]]
    name: str

    def validate_profile(
        self, dataset_profile: DatasetProfileView, reference_profile: DatasetProfileView
    ) -> Tuple[bool, Optional[Dict[str, Any]]]:
        validate_result, summary = self.condition(dataset_profile, reference_profile)
        return (validate_result, summary)


@dataclass(frozen=True)
class DatasetConstraint:
    """
    Implements dataset-level constraints that are not attached to a specific
    metric or column. The condition Callable takes the DatasetProfileView as
    input and returns a boolean indicating whether the condition is satisfied
    as well as a dictionary mapping 'column_name/metric_namespace' -> Metric for
    any metrics the condition used during its evaluation.
    """

    condition: Callable[[DatasetProfileView], Tuple[bool, Dict[str, Metric]]]
    name: str
    require_column_existence: bool = True  # Applies to all columns referenced in the constraint

    def _get_metric_summary(self, metrics: Dict[str, Metric]) -> Optional[Dict[str, Any]]:
        summary: Dict[str, Any] = dict()
        for path, metric in metrics.items():
            metric_summary = pretty_display(metric.namespace, metric.to_summary_dict())
            for component, value in metric_summary.items():
                summary[f"{path}/{component}"] = value
        return summary

    def validate_profile(self, dataset_profile: DatasetProfileView) -> Tuple[bool, Optional[Dict[str, Any]]]:
        try:
            validate_result, metrics = self.condition(dataset_profile)
        except MissingMetric as e:
            if self.require_column_existence:
                logger.info(f"validate_profile could not get metric {str(e)} so returning False.")
                return (False, None)

            logger.info(
                f"validate_profile could not get metric {str(e)}  "
                f"but require_column_existence is False, so returning True. "
            )
            return (True, None)

        metric_summary = self._get_metric_summary(metrics)
        return (validate_result, metric_summary)


class MetricConstraintWrapper:
    """
    Converts a MetricConstraint not associated with a specific column
    into a Callable that can be a DatasetConstraint condition.
    """

    def __init__(self, constraint: MetricConstraint) -> None:
        self.constraint = constraint

    def __call__(self, profile: DatasetProfileView) -> Tuple[bool, Dict[str, Metric]]:
        valid, _ = self.constraint.validate_profile(profile)
        metrics = self.constraint.metric_selector.apply(profile)  # Metrics used during evaluation
        columns = profile.get_columns()
        metric_map: Dict[str, Metric] = dict()
        # Find the columns the referenced Metrics live in
        for metric in metrics:
            for column_name, column_view in columns.items():
                if metric in column_view.get_metrics():
                    metric_map[f"{column_name}/{metric.namespace}"] = metric
                    break
        return valid, metric_map


_INT_RE = re.compile(r"[-+]?\d+")
_FLOAT_RE = re.compile(r"[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?")


class PrefixCondition:
    """
    Interpret expressions in the form used to serialize Predicate expressions.
    This is probably a reasonable serialization format for DatasetConstraint
    conditions, but might not be the best user interface for creating conditions
    in client code.
    """

    def __init__(self, expression: str, cfg: Optional[SummaryConfig] = None) -> None:
        self._expression = expression
        self._tokens = _tokenize(expression)
        self._profile = None
        self._metric_map: Dict[str, Metric] = dict()
        self._cfg = cfg or SummaryConfig()

    def _interpret(self, i: int) -> Tuple[Any, int]:  # noqa: C901
        token = self._tokens[i]

        # Metric reference        :column_name:metric_namespace/component_name
        # MultiMetric reference   :column_name:metric_namespace/submetric_name:submetric_namespace/component_name
        # Dataset-level metric    ::metric_namespace/...

        match = _METRIC_REF.fullmatch(token)
        if bool(match):
            path = token[2:]
            # TODO: get dataset-level metric from profile
            # TODO: self._metric_map[path] = metric
            return 0, i + 1

        match = _PROFILE_REF.fullmatch(token)
        if bool(match):
            column_name, path = match.groups()
            metric_name, component_name = path.split("/", 1)
            try:
                metric = self._profile.get_column(column_name).get_metric(metric_name)  # type: ignore
                summary = metric.to_summary_dict(self._cfg)
            except:  # noqa
                raise MissingMetric(token)

            # Track Metrics referenced during evaluation
            metric_path = f"{column_name}/{metric_name}"
            self._metric_map[metric_path] = metric
            try:
                value = summary[component_name]
            except:  # noqa
                raise ValueError(f"Component {component_name} not found in {metric_path}")

            return value, i + 1

        # Relational operators

        if token == "~":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            regex = re.compile(right) if isinstance(right, str) else None
            # TODO: should we try to str(left) ?
            return (regex and isinstance(left, str) and bool(regex.match(left))), i

        if token == "~=":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            regex = re.compile(right) if isinstance(right, str) else None
            # TODO: should we try to str(left) ?
            return (regex and isinstance(left, str) and bool(regex.fullmatch(left))), i

        if token == "==":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left == right), i

        if token == "!=":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left != right), i

        if token == "<":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left < right), i

        if token == "<=":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left <= right), i

        if token == ">":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left > right), i

        if token == ">=":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left >= right), i

        # Boolean operators

        if token == "and":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left and right), i

        if token == "or":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left or right), i

        if token == "not":
            right, i = self._interpret(i + 1)
            return (not right), i

        # Arithmetic operators

        if token == "+":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left + right), i

        if token == "-":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left - right), i

        if token == "*":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left * right), i

        if token == "**":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left**right), i

        if token == "/":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left / right), i

        if token == "//":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left // right), i

        if token == "%":
            left, i = self._interpret(i + 1)
            right, i = self._interpret(i)
            return (left % right), i

        # Literals

        if token.startswith('"'):
            return token[1:-1], i + 1

        if bool(_INT_RE.fullmatch(token)):
            return int(token), i + 1

        if bool(_FLOAT_RE.fullmatch(token)):
            return float(token), i + 1

        raise ValueError(f"Unparsable expression: '{self._expression}' at token {i}: '{token}'")

    def __call__(self, profile: DatasetProfileView) -> Tuple[bool, Optional[Dict[str, Metric]]]:
        """
        relational operators: ~ ~= == < <= > >= !=
        boolean operators:    and or not
        arithmetic operators: + - * ** / // %

        literals: 42 3.14 "blah"

        metric component references: :column_name:metric_namespace/metric_component

        returns a bool indicating whether the expression passed and a map of
        :column_name:metric_namespace -> Metric for the metrics used in evaluating
        the expression.
        """
        self._profile = profile
        self._metric_map = dict()
        passes, _ = self._interpret(0)
        if not isinstance(passes, bool):
            raise ValueError(f"Contraint expression should return a boolean, got {type(passes)}")

        return passes, self._metric_map


def _make_report(
    constraint_name: str, result: bool, with_summary: bool, metric_summary: Optional[Dict[str, Any]]
) -> ReportResult:
    if not result:
        if with_summary:
            return ReportResult(name=constraint_name, passed=0, failed=1, summary=metric_summary)
        return ReportResult(name=constraint_name, passed=0, failed=1, summary=None)

    if with_summary:
        return ReportResult(name=constraint_name, passed=1, failed=0, summary=metric_summary)
    return ReportResult(name=constraint_name, passed=1, failed=0, summary=None)


class Constraints:
    column_constraints: Dict[str, Dict[str, MetricConstraint]]
    dataset_constraints: List[DatasetConstraint]
    dataset_comparison_constraints: List[DatasetComparisonConstraint]
    dataset_profile_view: Optional[DatasetProfileView]

    def __init__(
        self,
        dataset_profile_view: Optional[DatasetProfileView] = None,
        column_constraints: Optional[Dict[str, Dict[str, MetricConstraint]]] = None,
        reference_profile_view: Optional[DatasetProfileView] = None,
    ) -> None:
        self.column_constraints = column_constraints or dict()
        self.dataset_constraints = []
        self.dataset_comparison_constraints = []
        self.dataset_profile_view = dataset_profile_view
        self.reference_profile_view = reference_profile_view

    def validate(self, profile_view: Optional[DatasetProfileView] = None) -> bool:
        profile = self._resolve_profile_view(profile_view)
        column_names = self.column_constraints.keys()
        if (
            len(column_names) == 0
            and len(self.dataset_constraints) == 0
            and len(self.dataset_comparison_constraints) == 0
        ):
            logger.warning("validate was called with empty set of constraints, returning True!")
            return True

        for column_name in column_names:
            columnar_constraints = self.column_constraints[column_name]
            for constraint_name, metric_constraint in columnar_constraints.items():
                (result, _) = metric_constraint.validate_profile(profile)
                if not result:
                    logger.info(f"{constraint_name} failed on column {column_name}")
                    return False

        for constraint in self.dataset_constraints + self.dataset_comparison_constraints:
            if isinstance(constraint, DatasetConstraint):
                (result, _) = constraint.validate_profile(profile)
            elif isinstance(constraint, DatasetComparisonConstraint):
                (result, _) = constraint.validate_profile(profile, self.reference_profile_view)
            if not result:
                logger.info(f"{constraint.name} failed on dataset")
                return False
        return True

    def set_reference_profile_view(self, profile_view: DatasetProfileView) -> None:
        self.reference_profile_view = profile_view

    @deprecated(message="Please use generate_constraints_report()")
    def report(self, profile_view: Optional[DatasetProfileView] = None) -> List[Tuple[str, int, int]]:
        profile = self._resolve_profile_view(profile_view)
        column_names = self.column_constraints.keys()
        results: List[Tuple[str, int, int]] = []
        for column_name in column_names:
            columnar_constraints = self.column_constraints[column_name]
            for constraint_name, metric_constraint in columnar_constraints.items():
                (result, _) = metric_constraint.validate_profile(profile)
                if not result:
                    logger.info(f"{constraint_name} failed on column {column_name}")
                    results.append((constraint_name, 0, 1))
                else:
                    results.append((constraint_name, 1, 0))
        for constraint in self.dataset_constraints:
            result, _ = constraint.validate_profile(profile)
            if not result:
                logger.info(f"{constraint.name} failed on dataset")
                results.append((constraint.name, 0, 1))
            else:
                results.append((constraint.name, 1, 0))

        if len(results) == 0:
            logger.warning("report was called with empty set of constraints!")

        return results

    def _generate_metric_report(
        self,
        profile_view: DatasetProfileView,
        metric_constraint: MetricConstraint,
        constraint_name: str,
        with_summary: bool,
    ) -> ReportResult:
        (result, metric_summary) = metric_constraint.validate_profile(profile_view)
        return _make_report(constraint_name, result, with_summary, metric_summary)

    def _generate_dataset_report(
        self,
        profile_view: DatasetProfileView,
        constraint: Union[DatasetConstraint, DatasetComparisonConstraint],
        with_summary: bool,
    ) -> ReportResult:
        if isinstance(constraint, DatasetComparisonConstraint):
            (result, summary) = constraint.validate_profile(profile_view, self.reference_profile_view)
        elif isinstance(constraint, DatasetConstraint):
            (result, summary) = constraint.validate_profile(profile_view)
        return _make_report(constraint.name, result, with_summary, summary)

    def generate_constraints_report(
        self, profile_view: Optional[DatasetProfileView] = None, with_summary=False
    ) -> List[ReportResult]:
        profile = self._resolve_profile_view(profile_view)
        column_names = self.column_constraints.keys()
        results: List[ReportResult] = []

        for column_name in column_names:
            columnar_constraints = self.column_constraints[column_name]
            for constraint_name, metric_constraint in columnar_constraints.items():
                metric_report = self._generate_metric_report(
                    profile_view=profile,
                    metric_constraint=metric_constraint,
                    constraint_name=constraint_name,
                    with_summary=with_summary,
                )
                if metric_report[1] == 0:
                    logger.info(f"{constraint_name} failed on column {column_name}")

                results.append(metric_report)

        for constraint in self.dataset_constraints + self.dataset_comparison_constraints:
            metric_report = self._generate_dataset_report(
                profile_view=profile,
                constraint=constraint,
                with_summary=with_summary,
            )
            if metric_report[1] == 0:
                logger.info(f"{constraint.name} failed on dataset.")

            results.append(metric_report)

        if len(results) == 0:
            logger.warning("generate_constraints_report was called with empty set of constraints!")

        return results

    def _resolve_profile_view(self, profile_view: Optional[DatasetProfileView]) -> DatasetProfileView:
        if profile_view is None and self.dataset_profile_view is None:
            raise ValueError(
                "Constraints need to be initialized on a profile or have a profile passed in before calling validate or report."
            )
        return profile_view if profile_view is not None else self.dataset_profile_view


class ConstraintsBuilder:
    def __init__(
        self,
        dataset_profile_view: DatasetProfileView,
        constraints: Optional[Constraints] = None,
        reference_profile_view: Optional[DatasetProfileView] = None,
    ) -> None:
        self._dataset_profile_view = dataset_profile_view
        if constraints is None:
            constraints = Constraints(dataset_profile_view=dataset_profile_view)
        self._constraints = constraints
        self._reference_profile_view = reference_profile_view

    def get_metric_selectors(self) -> List[MetricsSelector]:
        selectors = []
        column_profiles = self._dataset_profile_view.get_columns()
        for column_name, column_profile in column_profiles.items():
            metric_component_paths = column_profile._metrics.keys()
            for metric_path in metric_component_paths:
                selectors.append(MetricsSelector(metric_name=metric_path, column_name=column_name))
        return selectors

    def add_constraints(self, constraints: List[Union[MetricConstraint, DatasetConstraint]]) -> "ConstraintsBuilder":
        for constraint in constraints:
            self.add_constraint(constraint)
        return self

    def add_constraint(
        self, constraint: Union[MetricConstraint, DatasetConstraint], ignore_missing: bool = False
    ) -> "ConstraintsBuilder":
        if isinstance(constraint, MetricConstraint):
            # check that the column exists and we have a column profile view
            column_name = constraint.metric_selector.column_name
            column_profile_view = self._dataset_profile_view.get_column(column_name)
            if (
                column_name is not None
                and column_profile_view is None
                and not ignore_missing
                and constraint.metric_selector.metrics_resolver is None
            ):
                raise ValueError(
                    f"Column '{column_name}' was not found in this profile, the available columns are: {self._dataset_profile_view.get_columns()}"
                )

            metric_selector = constraint.metric_selector
            metrics = metric_selector.apply(self._dataset_profile_view)
            if (metrics is None or len(metrics) == 0) and not (ignore_missing or column_name is None):
                logger.warning(
                    f"metrics not found for {column_name}, available metric components are: {column_profile_view.get_metric_component_paths()}. Skipping {constraint.name}."  # noqa: E501
                )
                return self

            if column_name is None:
                # MetricConstraint not associated with a specific column is turned into
                # a DatasetConstraint. It relies on the constraint's metric_selector
                # to find the metrics it needs to evaluate the condition.
                self._constraints.dataset_constraints.append(
                    DatasetConstraint(
                        condition=MetricConstraintWrapper(constraint),
                        name=constraint.name,
                        require_column_existence=constraint.require_column_existence,
                    )
                )
            else:
                if column_name not in self._constraints.column_constraints:
                    self._constraints.column_constraints[column_name] = dict()
                self._constraints.column_constraints[column_name][constraint.name] = constraint

        if isinstance(constraint, DatasetConstraint):
            self._constraints.dataset_constraints.append(constraint)
        if isinstance(constraint, DatasetComparisonConstraint):
            if not self._reference_profile_view:
                raise ValueError(
                    "ConstraintsBuilder: reference_profile_view must be set when adding DatasetComparisonConstraint"
                )
            self._constraints.set_reference_profile_view(self._reference_profile_view)
            self._constraints.dataset_comparison_constraints.append(constraint)

        return self

    def build(self) -> Constraints:
        return deepcopy(self._constraints)



from itertools import takewhile
from typing import Any, Generator, List, Tuple, Type


def get_like_items(items: List[Any]) -> Tuple[List[Any], Type, List[Any]]:
    if not items:
        return ([], type(None), items)

    item_type = type(items[0])
    matches = list(takewhile(lambda item: isinstance(item, item_type), items))
    return (matches, item_type, items[len(matches) :])


def type_batched_items(items: List[Any]) -> Generator[Tuple[List[Any], Type], None, None]:
    (matches, item_type, rest) = get_like_items(items)
    yield (matches, item_type)
    if not rest:
        return
    else:
        yield from type_batched_items(rest)



import logging
from collections import defaultdict
from dataclasses import dataclass, field
from itertools import chain
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from whylogs.core.datatypes import (
    AnyType,
    DataType,
    Fractional,
    Integral,
    StandardTypeMapper,
    String,
    TypeMapper,
)
from whylogs.core.metrics import StandardMetric
from whylogs.core.metrics.metrics import (
    Metric,
    MetricConfig,
    OperationResult,
    register_metric,
)
from whylogs.core.metrics.multimetric import MultiMetric, SubmetricSchema
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.resolvers import (
    COLUMN_METRICS,
    DEFAULT_RESOLVER,
    DeclarativeResolverBase,
    MetricSpec,
    ResolverSpec,
)
from whylogs.core.schema import DeclarativeSchema
from whylogs.core.segmentation_partition import SegmentationPartition
from whylogs.core.utils import deprecated
from whylogs.core.validators.validator import Validator

logger = logging.getLogger(__name__)


class DeclarativeSubmetricSchema(DeclarativeResolverBase, SubmetricSchema):
    """
    The DeclarativeSubmetricSchema allows one to customize the set of metrics
    tracked for each UDF computed by a UdfMetric. Pass its constructor a list
    of ResolverSpecs, which specify the UDF name or data type to
    match and the list of MetricSpecs to instantiate for matching UDFs.
    Each MetricSpec specifies the Metric class and MetricConfig to
    instantiate. Omit the MetricSpec::config to use the default MetricConfig.
    Setting ResolverSpec::exclude to True will exclude the listed metrics from the matched UDFs.

    For example, DeclarativeSubmetricSchema(resolvers=STANDARD_RESOLVER) implements
    the same schema as DatasetSchema(), i.e., using the default MetricConfig,
    StandardTypeMapper, StandardResolver, etc.  STANDARD_RESOLVER is defined
    in whylogs/python/whylogs/core/resolvers.py
    """

    def __init__(self, resolvers: List[ResolverSpec], default_config: Optional[MetricConfig] = None) -> None:
        super().__init__(resolvers, default_config)
        for resolver in resolvers:
            for metric_spec in resolver.metrics:
                if issubclass(metric_spec.metric, MultiMetric) and not resolver.exclude:
                    raise ValueError(
                        f"MultiMetric cannot contain another MultiMetric ({metric_spec.metric.get_namespace()}) as a submetric"
                    )

    def resolve(self, name: str, why_type: DataType, fi_disabled: bool = False) -> Dict[str, Metric]:
        return self._resolve(name, why_type, None)


STANDARD_UDF_RESOLVER: List[ResolverSpec] = [
    ResolverSpec(
        column_type=Integral,
        metrics=COLUMN_METRICS
        + [
            MetricSpec(StandardMetric.distribution.value),
            MetricSpec(StandardMetric.ints.value),
            MetricSpec(StandardMetric.cardinality.value),
            MetricSpec(StandardMetric.frequent_items.value),
        ],
    ),
    ResolverSpec(
        column_type=Fractional,
        metrics=COLUMN_METRICS
        + [
            MetricSpec(StandardMetric.distribution.value),
            MetricSpec(StandardMetric.cardinality.value),
        ],
    ),
    ResolverSpec(
        column_type=String,
        metrics=COLUMN_METRICS
        + [
            MetricSpec(StandardMetric.distribution.value),
            MetricSpec(StandardMetric.cardinality.value),
            MetricSpec(StandardMetric.frequent_items.value),
        ],
    ),
    ResolverSpec(column_type=AnyType, metrics=COLUMN_METRICS),
]

DEFAULT_UDF_RESOLVER: List[ResolverSpec] = STANDARD_UDF_RESOLVER


def default_schema() -> DeclarativeSubmetricSchema:
    return DeclarativeSubmetricSchema(DEFAULT_UDF_RESOLVER)


@dataclass(frozen=True)
class UdfMetricConfig(MetricConfig):
    """
    Configure UDFs & submetrics for UdfMetric

    Attributes:
       udfs: Maps submetric name to the UDF that computes the value to track

       submetric_schema [optional]: determines the set of metrics tracked for
           each computed value

       type_mapper [optional]: maps Python types to whylogs DataType
    """

    udfs: Dict[str, Callable[[Any], Any]] = field(default_factory=dict)
    submetric_schema: SubmetricSchema = field(default_factory=default_schema)
    type_mapper: TypeMapper = field(default_factory=StandardTypeMapper)


class UdfMetric(MultiMetric):
    """
    Applies the specified UDFs to the input column values and tracks the metrics
    specified by the submetric_schema to their output.

    Args:
        udfs: map of submetric name to UDF to compute value to track

        submetric_schema [optional]: determines the set of metrics to track for
            each UDF output. Defaults to STANDARD_RESOLVER

        type_mapper [optional]: maps Python types to whylogs DataType. Defaults
            to StandardTypeMapper

        fi_disabled [optional]: Should FrequentItemsMetric tracking be disabled.
            Defaults to False
    """

    def __init__(
        self,
        submetrics: Dict[str, Dict[str, Metric]],
        udfs: Optional[Dict[str, Callable[[Any], Any]]] = None,
        # discover these with resolver  submetrics: Dict[str, Dict[str, Metric]],  # feature name -> (namespace -> metric)
        submetric_schema: Optional[SubmetricSchema] = None,
        type_mapper: Optional[TypeMapper] = None,
        fi_disabled: bool = False,
    ):
        super().__init__(submetrics)
        self._udfs = udfs or dict()
        self._submetric_schema = submetric_schema or default_schema()
        self._type_mapper = type_mapper or StandardTypeMapper()
        self._fi_disabled = fi_disabled

    @property
    def namespace(self) -> str:
        return "udf"

    def merge(self, other: "UdfMetric") -> "UdfMetric":
        merged = UdfMetric(self.submetrics, self._udfs, self._submetric_schema, self._type_mapper, self._fi_disabled)
        merged.submetrics = self.merge_submetrics(other)
        return merged

    def _add_submetric(self, name: str, value: Any) -> None:
        self.submetrics[name] = self._submetric_schema.resolve(name, self._type_mapper(type(value)), self._fi_disabled)

    def _update_relevant_submetrics(self, name: str, data: PreprocessedColumn) -> None:
        if name not in self.submetrics:
            return

        for metric in self.submetrics[name].values():
            metric.columnar_update(data)

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        successes = 0
        failures = 0
        for value in list(chain.from_iterable(view.raw_iterator())):
            ok = True
            for submetric_name, udf in self._udfs.items():
                try:
                    computed_value = udf(value)
                    if submetric_name not in self.submetrics:
                        self._add_submetric(submetric_name, computed_value)  # NOTE: assumes column is homogeneous-ish?

                    data = PreprocessedColumn._process_scalar_value(computed_value)
                    self._update_relevant_submetrics(submetric_name, data)
                except Exception:  # noqa
                    logger.exception(f"UDF {submetric_name} evaluation failed")
                    ok = False

            if ok:
                successes += 1
            else:
                failures += 1

        return OperationResult(failures, successes)

    @classmethod
    def zero(cls, config: Optional[MetricConfig] = None) -> "UdfMetric":
        config = config or UdfMetricConfig()
        if not isinstance(config, UdfMetricConfig):
            logger.error("UdfMetric.zero() needs an UdfMetricConfig")
            config = UdfMetricConfig()

        return UdfMetric(
            dict(),
            udfs=config.udfs,
            submetric_schema=config.submetric_schema,
            type_mapper=config.type_mapper,
            fi_disabled=config.fi_disabled,
        )


# Register it so Multimetric and ProfileView can deserialize
register_metric(UdfMetric)


# The following variables hold information about the registered UDFs used to generate
# the resolvers/schemas. They are indexed by schema name and column name[type],
# and collect the registered information.

# _col_name_submetrics[schema_name][column_name] -> list (submetric_name, UDF)
_col_name_submetrics: Dict[str, Dict[str, List[Tuple[str, Callable[[Any], Any]]]]] = defaultdict(
    lambda: defaultdict(list)
)
_col_name_submetric_schema: Dict[str, Dict[str, SubmetricSchema]] = defaultdict(dict)
_col_name_type_mapper: Dict[str, Dict[str, TypeMapper]] = defaultdict(dict)

_col_type_submetrics: Dict[str, Dict[DataType, List[Tuple[str, Callable[[Any], Any]]]]] = defaultdict(
    lambda: defaultdict(list)
)
_col_type_submetric_schema: Dict[str, Dict[DataType, SubmetricSchema]] = defaultdict(dict)
_col_type_type_mapper: Dict[str, Dict[DataType, TypeMapper]] = defaultdict(dict)


def _reset_metric_udfs() -> None:
    global _col_name_submetrics, _col_name_submetric_schema, _col_name_type_mapper
    global _col_type_submetrics, _col_type_submetric_schema, _col_type_type_mapper
    _col_name_submetrics = defaultdict(lambda: defaultdict(list))
    _col_name_submetric_schema = defaultdict(dict)
    _col_name_type_mapper = defaultdict(dict)
    _col_type_submetrics = defaultdict(lambda: defaultdict(list))
    _col_type_submetric_schema = defaultdict(dict)
    _col_type_type_mapper = defaultdict(dict)


def register_metric_udf(
    col_name: Optional[str] = None,
    col_type: Optional[DataType] = None,
    submetric_name: Optional[str] = None,
    submetric_schema: Optional[SubmetricSchema] = None,
    type_mapper: Optional[TypeMapper] = None,
    namespace: Optional[str] = None,
    schema_name: str = "",
) -> Callable[[Any], Any]:
    """
    Decorator to easily configure UdfMetrics for your data set. Decorate your UDF
    functions, then call generate_udf_schema() to generate a list of ResolverSpecs
    that include the UdfMetrics configured by your decorator parameters.

    You must specify exactly one of either col_name or col_type. col_name will attach
    a UdfMetric to the named input column. col_type will attach a UdfMetric to all
    input columns of the specified type. The decorated function will automatically
    be a UDF in the UdfMetric.

    Specify submetric_name to give the output of the UDF a name. submetric_name
    defautls to the name of the decorated function. Note that all lambdas are
    named "lambda" so omitting submetric_name on more than one lambda will result
    in name collisions. If you pass a namespace, it will be prepended to the UDF name.

    You can optionally pass submetric_schema to specify and configure the metrics
    to be tracked for each UDF. This defualts to the STANDARD_RESOLVER metrics.

    You can optionally pass type_mapper to control how Python types are mapped to
    whylogs DataTypes. This defaults to the StandardTypeMapper.
    """

    def decorator_register(func):
        global _col_name_submetrics, _col_name_submetric_schema, _col_name_type_mapper
        global _col_type_submetrics, _col_type_submetric_schema, _col_type_type_mapper

        if col_name is not None and col_type is not None:
            raise ValueError("Only specify one of column name or type")

        if col_name is None and col_type is None:
            raise ValueError("Must specify one of column name or type")

        subname = submetric_name or func.__name__
        subname = f"{namespace}.{subname}" if namespace else subname
        if col_name is not None:
            _col_name_submetrics[schema_name][col_name].append((subname, func))
            if submetric_schema is not None:
                if col_name in _col_name_submetric_schema[schema_name]:
                    logger.warn(f"Overwriting submetric schema for column {col_name}")
                _col_name_submetric_schema[schema_name][col_name] = submetric_schema
            if type_mapper is not None:
                if col_name in _col_name_type_mapper[schema_name]:
                    logger.warn(f"Overwriting UdfMetric type mapper for column {col_name}")
                _col_name_type_mapper[schema_name][col_name] = type_mapper
        else:
            _col_type_submetrics[schema_name][col_type].append((subname, func))
            if submetric_schema is not None:
                if col_type in _col_type_submetric_schema[schema_name]:
                    logger.warn(f"Overwriting submetric schema for column type {col_type}")
                _col_type_submetric_schema[schema_name][col_type] = submetric_schema
            if type_mapper is not None:
                if col_type in _col_type_type_mapper[schema_name]:
                    logger.warn(f"Overwriting UdfMetric type mapper for column type {col_type}")
                _col_type_type_mapper[schema_name][col_type] = type_mapper

        return func

    return decorator_register


def generate_udf_resolvers(
    schema_name: Union[str, List[str]] = "",
    include_default_schema: bool = True,
) -> List[ResolverSpec]:
    """
    Generates a list of ResolverSpecs that implement the UdfMetrics specified
    by the @register_metric_udf decorators. The result only includes the UdfMetric,
    so you may want to append it to a list of ResolverSpecs defining the other
    metrics you wish to track.

    For example:

    @register_metric_udf(col_name="col1")
    def add5(x):
        return x + 5

    @register_metric_udf(col_type=String)
    def upper(x):
        return x.upper()

    schema = DeclarativeSchema(STANDARD_RESOLVER + generate_udf_schema())
    why.log(data, schema=schema)

    This will attach a UdfMetric to column "col1" that will include a submetric
    named "add5" tracking the values in "col1" incremented by 5, and a UdfMetric
    for each string column that will include a submetric named "upper" tracking
    the uppercased strings in the input columns. Since these are appended to the
    STANDARD_RESOLVER, the default metrics are also tracked for every column.
    """

    resolvers: List[ResolverSpec] = list()
    udfs: Dict[str, Callable[[Any], Any]]
    schema_names = schema_name if isinstance(schema_name, list) else [schema_name]
    if include_default_schema and "" not in schema_names:
        schema_names = [""] + schema_names

    col_udfs: Dict[str, List[Tuple[str, Callable[[Any], Any]]]] = defaultdict(list)
    col_submetric_schema: Dict[str, SubmetricSchema] = dict()
    col_type_mapper: Dict[str, TypeMapper] = dict()

    type_udfs: Dict[DataType, List[Tuple[str, Callable[[Any], Any]]]] = defaultdict(list)
    type_submetric_schema: Dict[DataType, SubmetricSchema] = dict()
    type_type_mapper: Dict[DataType, TypeMapper] = dict()

    for schema_name in schema_names:
        for col_name, submetrics in _col_name_submetrics[schema_name].items():
            col_udfs[col_name] += submetrics
            if col_submetric_schema.get(col_name) and _col_name_submetric_schema[schema_name].get(col_name):
                logger.warning(
                    f"Multiple submetric schemas registered for column {col_name}. Which one will be used is undefined."
                )

            col_submetric_schema[col_name] = col_submetric_schema.get(col_name) or _col_name_submetric_schema[
                schema_name
            ].get(col_name)
            if col_type_mapper.get(col_name) and _col_name_type_mapper[schema_name].get(col_name):
                logger.warning(
                    f"Multiple submetric type mappers registerd for column {col_name}. Which one will be used is undefined."
                )

            col_type_mapper[col_name] = col_type_mapper.get(col_name) or _col_name_type_mapper[schema_name].get(
                col_name
            )

        for col_type, submetrics in _col_type_submetrics[schema_name].items():
            type_udfs[col_type] += submetrics
            if type_submetric_schema.get(col_type) and _col_type_submetric_schema[schema_name].get(col_type):
                logger.warning(
                    f"Multiple submetric schemas registered for column type {col_type}. Which one will be used is undefined."
                )

            type_submetric_schema[col_type] = type_submetric_schema.get(col_type) or _col_type_submetric_schema[
                schema_name
            ].get(col_type)
            if type_type_mapper.get(col_type) and _col_type_type_mapper[schema_name].get(col_type):
                logger.warning(
                    f"Multiple submetric type mappers registerd for column type {col_type}. Which one will be used is undefined."
                )

            type_type_mapper[col_type] = type_type_mapper.get(col_type) or _col_type_type_mapper[schema_name].get(
                col_type
            )

    for col_name, submetrics in col_udfs.items():
        udfs = dict()
        for submetric in submetrics:
            udfs[submetric[0]] = submetric[1]
        config = UdfMetricConfig(
            udfs=udfs,
            submetric_schema=col_submetric_schema.get(col_name) or default_schema(),
            type_mapper=col_type_mapper.get(col_name) or StandardTypeMapper(),
        )
        resolvers.append(ResolverSpec(col_name, None, [MetricSpec(UdfMetric, config)]))

    for col_type, submetrics in type_udfs.items():
        udfs = dict()
        for submetric in submetrics:
            udfs[submetric[0]] = submetric[1]
        config = UdfMetricConfig(
            udfs=udfs,
            submetric_schema=type_submetric_schema.get(col_type) or default_schema(),
            type_mapper=type_type_mapper.get(col_type) or StandardTypeMapper(),
        )
        resolvers.append(ResolverSpec(None, col_type, [MetricSpec(UdfMetric, config)]))

    return resolvers


@deprecated(message="Please use generate_udf_resolvers()")
def generate_udf_schema() -> List[ResolverSpec]:
    return generate_udf_resolvers()


def udf_metric_schema(
    non_udf_resolvers: Optional[List[ResolverSpec]] = None,
    types: Optional[Dict[str, Any]] = None,
    default_config: Optional[MetricConfig] = None,
    type_mapper: Optional[TypeMapper] = None,
    cache_size: int = 1024,
    schema_based_automerge: bool = False,
    segments: Optional[Dict[str, SegmentationPartition]] = None,
    validators: Optional[Dict[str, List[Validator]]] = None,
    schema_name: Union[str, List[str]] = "",
    include_default_schema: bool = True,
) -> DeclarativeSchema:
    """
    Generates a DeclarativeSchema that implement the UdfMetrics specified
    by the @register_metric_udf decorators (in additon to any non_udf_resolvers
    passed in).

    For example:

    @register_metric_udf(col_name="col1")
    def add5(x):
        return x + 5

    @register_metric_udf(col_type=String)
    def upper(x):
        return x.upper()

    why.log(data, schema=udf_metric_schema())

    This will attach a UdfMetric to column "col1" that will include a submetric
    named "add5" tracking the values in "col1" incremented by 5, and a UdfMetric
    for each string column that will include a submetric named "upper" tracking
    the uppercased strings in the input columns. Since these are appended to the
    STANDARD_RESOLVER, the default metrics are also tracked for every column.
    """

    resolvers = generate_udf_resolvers(schema_name, include_default_schema)
    non_udf_resolvers = non_udf_resolvers if non_udf_resolvers is not None else DEFAULT_RESOLVER

    return DeclarativeSchema(
        non_udf_resolvers + resolvers,
        types,
        default_config,
        type_mapper,
        cache_size,
        schema_based_automerge,
        segments,
        validators,
    )



from typing import List

from whylogs.core.constraints.factories import column_has_zero_count_types
from whylogs.core.constraints.metric_constraints import MetricConstraint
from whylogs.core.view.column_profile_view import ColumnProfileView


def generate_column_types_constraints(column_name: str, column_profile: ColumnProfileView) -> List[MetricConstraint]:
    """
    Generates constraints for the types metrics of a column.
    Parameters
    ----------
    column_name : str
        Name of the column
    column_profile : ColumnProfileView
        Profile of the column
    """
    constraints = []
    types_metric = column_profile.get_metric("types")
    zero_empty_types = []
    for key, component in vars(types_metric).items():
        if component.value == 0:
            zero_empty_types.append(key)
    if zero_empty_types:
        constraints.append(column_has_zero_count_types(column_name=column_name, types_list=zero_empty_types))
    return constraints



"""
A read/write library for length-delimited protobuf messages.

Based on: https://github.com/soulmachine/delimited-protobuf/blob/main/delimited_protobuf.py"""
import warnings
from logging import getLogger
from typing import IO, Type, TypeVar

from google.protobuf.message import Message

from whylogs.core.errors import DeserializationError

T = TypeVar("T", bound=Message)
logger = getLogger(__name__)
_DECODE_ERROR_STRING = "Unexpected end-group tag"


def _read_varint(stream: IO[bytes], offset: int = 0) -> int:
    from google.protobuf.internal.decoder import _DecodeVarint  # type: ignore

    """Read a varint from the stream."""
    if offset > 0:
        stream.seek(offset)
    buf: bytes = stream.read(1)
    if buf == b"":
        return 0  # reached EOF
    while (buf[-1] & 0x80) >> 7 == 1:  # while the MSB is 1
        new_byte = stream.read(1)
        if new_byte == b"":
            raise EOFError("unexpected EOF")
        buf += new_byte
    varint, _ = _DecodeVarint(buf, 0)
    return varint


def read_delimited_protobuf(stream: IO[bytes], proto_class_name: Type[T], offset: int = 0) -> T:
    """Read a single length-delimited message from the given stream."""
    size = _read_varint(stream, offset=offset)
    if size == 0:
        return proto_class_name()
    buf = stream.read(size)
    msg = proto_class_name()
    delimited_read_failure = False

    with warnings.catch_warnings(record=True) as warning_messages:
        try:
            logger.info(f"Attempting to read profile file as delimited from position {offset}.")
            msg.ParseFromString(buf)
        except Exception as decode_error:
            logger.warning(
                f"{decode_error}: when reading delimited protobuf file of size "
                f"{size} bytes as {proto_class_name}, falling back to non-delimited read."
            )
            delimited_read_failure = True

        if not delimited_read_failure and len(warning_messages) > 0:
            for w in warning_messages:
                if issubclass(w.category, RuntimeWarning):
                    logger.info(
                        f"Encountered RuntimeWarning when reading delimited file: {w.message}, retrying to read profile as non-delimited file."
                    )
                    if _DECODE_ERROR_STRING in str(w.message):
                        delimited_read_failure = True
                        break
            if not delimited_read_failure:
                logger.info(
                    f"Encountered {len(warning_messages)} warnings, e.g. {warning_messages[-1].message} but no DecodeError message detected, "
                    "proceeding with the already delimited parsed file."
                )

    if delimited_read_failure:
        logger.info("Attempting to read profile file as non-delimited from the beggining of the file.")
        stream.seek(0)
        buf = stream.read()
        try:
            msg.ParseFromString(buf)
            logger.info("successfully read profile file as non-delimited.")
        except Exception as e:
            logger.error(f"{e}: Occured during attempted fallback read of {proto_class_name}.")
            raise DeserializationError(f"Failed fallback attempt to read {proto_class_name}:{e}")

    return msg


def write_delimited_protobuf(stream: IO[bytes], msg: T) -> None:
    """Write a single length-delimited message to the given stream."""
    assert stream is not None
    from google.protobuf.internal.encoder import _EncodeVarint  # type: ignore

    _EncodeVarint(stream.write, msg.ByteSize())
    stream.write(msg.SerializeToString())



import re
from typing import List, Optional, Tuple

from whylogs.core.dataset_profile import DatasetProfile
from whylogs.core.metric_getters import MetricGetter, ProfileGetter
from whylogs.core.metrics.metrics import Metric
from whylogs.core.relations import (
    LiteralGetter,
    Predicate,
    Relation,
    ValueGetter,
    unescape_colon,
    unescape_quote,
)


def _unescape_token(input: str) -> str:
    if input[0] == '"':
        return unescape_quote(input)
    if input[0] == ":":
        return unescape_colon(input)
    return input


# We use this with findall() to split an input string into tokens.
# (?::(?:[^:]|\\:)*:[_a-zA-Z0-9]+/[_a-zA-Z0-9:/]+) matches metric references with \: escaped in column names
# "(?:[^\\"]|\\[^"]|\\")*" matches quoted strings with \"  \\ escaped quotes and \
# (?:[^ ":][^ ]*) matches everything else delimited by spaces

_TOKEN_RE = re.compile(r'(?::(?:[^:]|\\:)*:[_a-zA-Z0-9]+/[_a-zA-Z0-9:/]+)|(?:[^ ":][^ ]*)|"(?:[^\\"]|\\[^"]|\\")*"')


def _tokenize(input: str) -> List[str]:
    return [_unescape_token(t) for t in _TOKEN_RE.findall(input)]


def _get_component(token: List[str], i: int) -> Tuple[str, int]:
    return token[i], i + 1  # either dummy variable x or metric component name in metric::to_summary_dict()


# These match metric references with our without column names, with capture groups to
# pull out the column name and path. They have an optional submetric_name:submetric_namespace/
# section to support MultiMetric references.

_METRIC_REF = re.compile(r"::[_a-zA-Z0-9]+/(?:[_a-zA-Z0-9]+:[_a-zA-Z0-9]+/)?[_a-zA-Z0-9]+")
_PROFILE_REF = re.compile(r":(.+?):([_a-zA-Z0-9]+/(?:[_a-zA-Z0-9]+:[_a-zA-Z0-9]+/)?[_a-zA-Z0-9]+)")


def _get_value(
    token: List[str], i: int, metric: Optional[Metric] = None, profile: Optional[DatasetProfile] = None
) -> Tuple[ValueGetter, int]:
    if token[i].startswith('"'):
        return LiteralGetter(token[i][1:-1]), i + 1

    if _METRIC_REF.fullmatch(token[i]):
        if metric is None:
            raise ValueError("Must specify metric to use with MetricGetter")

        namespace, path = token[i][2:].split("/", 1)
        if metric.namespace != namespace:
            raise ValueError(f"Expected {namespace} metric but got {metric.namespace}")

        return MetricGetter(metric, path), i + 1

    match = _PROFILE_REF.fullmatch(token[i])
    if bool(match):
        if profile is None:
            raise ValueError("Must specify profile to use with ProfileGetter")

        column_name, path = match.groups()  # type: ignore
        return ProfileGetter(profile, column_name, path), i + 1

    if bool(re.fullmatch(r"[-+]?\d+", token[i])):
        return LiteralGetter(int(token[i])), i + 1

    return LiteralGetter(float(token[i])), i + 1


def _deserialize(
    token: List[str], i: int, metric: Optional[Metric] = None, profile: Optional[DatasetProfile] = None
) -> Tuple[Predicate, int]:
    if token[i] == "~":
        component, i = _get_component(token, i + 1)
        value, i = _get_value(token, i, metric, profile)
        return Predicate(Relation.match, value(), component=component), i

    if token[i] == "~=":
        component, i = _get_component(token, i + 1)
        value, i = _get_value(token, i, metric, profile)
        return Predicate(Relation.fullmatch, value(), component=component), i

    if token[i] == "==":
        component, i = _get_component(token, i + 1)
        value, i = _get_value(token, i, metric, profile)
        return Predicate(Relation.equal, value, component=component), i

    if token[i] == "<":
        component, i = _get_component(token, i + 1)
        value, i = _get_value(token, i, metric, profile)
        return Predicate(Relation.less, value, component=component), i

    if token[i] == "<=":
        component, i = _get_component(token, i + 1)
        value, i = _get_value(token, i, metric, profile)
        return Predicate(Relation.leq, value, component=component), i

    if token[i] == ">":
        component, i = _get_component(token, i + 1)
        value, i = _get_value(token, i, metric, profile)
        return Predicate(Relation.greater, value, component=component), i

    if token[i] == ">=":
        component, i = _get_component(token, i + 1)
        value, i = _get_value(token, i, metric, profile)
        return Predicate(Relation.geq, value, component=component), i

    if token[i] == "!=":
        component, i = _get_component(token, i + 1)
        value, i = _get_value(token, i, metric, profile)
        return Predicate(Relation.neq, value, component=component), i

    if token[i] == "and":
        left, i = _deserialize(token, i + 1)
        right, i = _deserialize(token, i, metric, profile)
        return Predicate(Relation._and, left=left, right=right, component=left._component), i

    if token[i] == "or":
        left, i = _deserialize(token, i + 1)
        right, i = _deserialize(token, i, metric, profile)
        return Predicate(Relation._or, left=left, right=right, component=left._component), i

    if token[i] == "not":
        right, i = _deserialize(token, i + 1)
        return Predicate(Relation._not, right=right, component=right._component), i

    raise ValueError("Unable to parse predicate expression '{' '.join(token)}' at token {i+1}.")


def parse_predicate(
    expression: str, metric: Optional[Metric] = None, profile: Optional[DatasetProfile] = None
) -> Predicate:
    predicate, _ = _deserialize(_tokenize(expression), 0, metric, profile)
    return predicate



from whylogs.api.usage_stats import emit_usage

from .metric_constraints import (
    Constraints,
    ConstraintsBuilder,
    DatasetComparisonConstraint,
    DatasetConstraint,
    MetricConstraint,
    MetricsSelector,
    PrefixCondition,
)

__ALL__ = [
    Constraints,
    ConstraintsBuilder,
    DatasetConstraint,
    DatasetComparisonConstraint,
    MetricConstraint,
    MetricsSelector,
    PrefixCondition,
]

emit_usage("metric_constraints")



import time
from datetime import datetime, timezone
from typing import Optional

from whylogs.core.utils.utils import ensure_timezone


def _get_event_file_name(
    trace_id: str,
    timestamp: Optional[datetime] = None,
    column_name: Optional[str] = None,
    prefix: str = "debug_event_",
    format: str = "%Y-%m-%d_%H-%M",
    file_extension: str = "json",
) -> str:
    if not timestamp:
        time_tuple = datetime.now(timezone.utc).utctimetuple()
    else:
        ensure_timezone(timestamp)
        time_tuple = timestamp.utctimetuple()

    feature_name = f"{column_name}_" if column_name else ""

    timed_filename = f"{prefix}{feature_name}{trace_id}.{time.strftime(format, time_tuple)}.{file_extension}"
    return timed_filename



from typing import List

from whylogs.core.constraints.factories import column_is_probably_unique
from whylogs.core.constraints.metric_constraints import MetricConstraint
from whylogs.core.utils import is_probably_unique
from whylogs.core.utils.stats_calculations import only_null_values
from whylogs.core.view.column_profile_view import ColumnProfileView


def generate_column_multi_metrics_constraints(
    column_name: str, column_profile: ColumnProfileView
) -> List[MetricConstraint]:
    constraints = []
    if is_probably_unique(column_profile) and not only_null_values(column_profile):
        constraints.append(column_is_probably_unique(column_name))
    return constraints



from io import BytesIO

import numpy as np

from whylogs.core.metrics.deserializers import deserializer
from whylogs.core.metrics.metric_components import MetricComponent
from whylogs.core.metrics.serializers import serializer
from whylogs.core.proto import MetricComponentMessage


def _serialize_ndarray(a: np.ndarray) -> bytes:
    bio = BytesIO()
    np.save(bio, a, allow_pickle=False)
    return bio.getvalue()


def _deserialize_ndarray(a: bytes) -> np.ndarray:
    bio = BytesIO(a)
    return np.load(bio, allow_pickle=False)


class MatrixComponent(MetricComponent[np.ndarray]):
    # mtype = np.ndarray
    type_id = 101


@serializer(type_id=101)
def serialize(value: np.ndarray) -> MetricComponentMessage:
    return MetricComponentMessage(serialized_bytes=_serialize_ndarray(value))


@deserializer(type_id=101)
def deserialize(msg: MetricComponentMessage) -> np.ndarray:
    return _deserialize_ndarray(msg.serialized_bytes)



from .protobuf_utils import read_delimited_protobuf, write_delimited_protobuf
from .stats_calculations import (
    get_cardinality_estimate,
    get_distribution_metrics,
    is_probably_unique,
)
from .utils import deprecated, deprecated_alias, ensure_timezone

__ALL__ = [  #
    read_delimited_protobuf,
    write_delimited_protobuf,
    get_distribution_metrics,
    is_probably_unique,
    get_cardinality_estimate,
    deprecated_alias,
    deprecated,
    ensure_timezone,  #
]



from concurrent.futures import Future, wait
from typing import Callable, Optional, TypeVar

T = TypeVar("T")


def _wait_result(future: "Future[T]", timeout: Optional[float] = None) -> T:
    """
    Wait on a future with an optional timeout without side effects. This won't update
    the status of the future for errors/timeouts.
    """
    done, not_done = wait([future], timeout=timeout)

    if len(not_done) > 0:
        raise TimeoutError("Timeout waiting for result")

    all = done.union(not_done)
    for it in all:
        e = it.exception()
        r = it.result()

        if e is not None:
            raise e
        elif it.cancelled():
            raise Exception("cancelled")
        else:
            return r

    raise Exception("Couldn't find a result")


def wait_result(future: "Future[T]", timeout: Optional[float] = None) -> T:
    """
    Wait on a future with an optional timeout.
    """
    try:
        return _wait_result(future, timeout=timeout)
    except TimeoutError as e:
        future.set_exception(e)
        raise e


def wait_result_while(future: "Future[T]", predicate: Callable[[], bool]) -> T:
    """
    Wait on a future while the condition is true.
    """
    try:
        while predicate():
            try:
                return _wait_result(future, 1.0)
            except TimeoutError:
                pass
    except Exception as e:
        if future.exception() is None:
            future.set_exception(e)
        raise e

    raise TimeoutError("Wait signal stopped before result was available.")



import atexit
import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from whylogs.api.logger.result_set import (
    ProfileResultSet,
    ResultSet,
    SegmentedResultSet,
)
from whylogs.api.logger.segment_processing import segment_processing
from whylogs.api.store import ProfileStore
from whylogs.api.writer import Writer, Writers
from whylogs.core import DatasetProfile, DatasetSchema
from whylogs.core.errors import LoggingError
from whylogs.core.input_resolver import _pandas_or_dict
from whylogs.core.metadata import (
    _populate_common_profile_metadata,
    _safe_merge_metadata,
)
from whylogs.core.stubs import pd

logger = logging.getLogger(__name__)


class BasicCache(object):
    _cache: Dict[DatasetSchema, DatasetProfile] = {}

    def get(self, schema: DatasetSchema) -> DatasetProfile:
        candidate = self._cache.get(schema)
        if candidate is None:
            candidate = DatasetProfile(schema=schema)
            self._cache[schema] = candidate
        return candidate


class Logger(ABC):
    def __init__(self, schema: Optional[DatasetSchema] = None):
        self._is_closed = False
        self._schema = schema
        self._writers: List[Writer] = []
        atexit.register(self.close)
        self._store_list: List[ProfileStore] = []
        self._segment_cache = None
        self._metadata: Optional[Dict[str, str]] = None

    def check_writer(self, _: Writer) -> None:
        """Checks if a writer is configured correctly for this class"""

    def append_writer(self, name: Optional[str] = None, *, writer: Optional[Writer] = None, **kwargs: Any) -> None:
        if name is None and writer is None:
            raise ValueError("Must specify either the writer name or a Writer object")
        if name is not None and writer is not None:
            raise ValueError("Cannot specify name and writer at the same time")
        if name is not None:
            writer = Writers.get(name, **kwargs)

        assert writer is not None
        self.check_writer(writer)
        self._writers.append(writer)

    def append_store(self, store: ProfileStore) -> None:
        self._store_list.append(store)

    @abstractmethod
    def _get_matching_profiles(
        self,
        obj: Any = None,
        *,
        pandas: Optional[pd.DataFrame] = None,
        row: Optional[Dict[str, Any]] = None,
        schema: Optional[DatasetSchema] = None,
    ) -> List[DatasetProfile]:
        pass

    def log(
        self,
        obj: Any = None,
        *,
        pandas: Optional[pd.DataFrame] = None,
        row: Optional[Dict[str, Any]] = None,
        schema: Optional[DatasetSchema] = None,
        timestamp_ms: Optional[int] = None,  # Not the dataset timestamp, but the timestamp of the data
        name: Optional[str] = None,
        trace_id: Optional[str] = None,
        tags: Optional[List[str]] = None,
        segment_key_values: Optional[Dict[str, str]] = None,
    ) -> ResultSet:
        """
        Args:
            timestamp_ms: The timestamp of the data being logged. This defaults to now if it isn't provided.
             This is used to determine what the dataset timestamp should be. For an hourly model, the dataset
             timestamp will end up being the start of the hour of the provided timestamp_ms, UTC.
        """
        if self._is_closed:
            raise LoggingError("Cannot log to a closed logger")
        if obj is None and pandas is None and row is None:
            # TODO: check for shell environment and emit more verbose error string to let user know how to correct.
            raise LoggingError("log() was called without passing in any input!")

        if name is not None:
            if self._metadata is None:
                self._metadata = dict()
            self._metadata["name"] = name
        active_schema = schema or self._schema
        if active_schema:
            pandas, row = _pandas_or_dict(obj, pandas, row)
            obj = None
            pandas, row = active_schema._run_udfs(pandas, row)

        # If segments are defined use segment_processing to return a SegmentedResultSet
        if active_schema and active_schema.segments:
            segmented_results: SegmentedResultSet = segment_processing(
                schema=active_schema,
                obj=obj,
                pandas=pandas,
                row=row,
                segment_cache=self._segment_cache,
            )
            # Update the existing segmented_results metadata with the trace_id and other keys if not present
            _populate_common_profile_metadata(segmented_results.metadata, trace_id=trace_id, tags=tags)
            _safe_merge_metadata(default_metadata=segmented_results.metadata, incoming_metadata=active_schema.metadata)
            return segmented_results

        profiles = self._get_matching_profiles(obj, pandas=pandas, row=row, schema=active_schema)

        for prof in profiles:
            prof.track(obj, pandas=pandas, row=row, execute_udfs=False)
            prof._metadata = _populate_common_profile_metadata(prof._metadata, trace_id=trace_id, tags=tags)
            if active_schema:
                _safe_merge_metadata(prof._metadata, active_schema.metadata)

        first_profile = profiles[0]
        if name is not None:
            if first_profile._metadata is None:
                first_profile._metadata = dict()
            first_profile._metadata["name"] = name
        return ProfileResultSet(first_profile)

    def close(self) -> None:
        self._is_closed = True

    def __enter__(self) -> "Logger":
        return self

    def __exit__(self, exception_type: Any, exception_value: Any, traceback: Any) -> None:
        if not self._is_closed:
            self.close()



def package_version(package: str = __package__) -> str:
    """Calculate version number based on pyproject.toml"""
    try:
        from importlib import metadata  # type: ignore
    except ImportError:
        # Python < 3.8
        import importlib_metadata as metadata  # type: ignore

    try:
        version = metadata.version(package)
    except metadata.PackageNotFoundError:  # type: ignore
        version = f"{package} is not installed."

    return version


whylogs_version = package_version("whylogs")
